import os
import pickle
import shutil
import warnings
from typing import Any, Dict

import torch
from torch_geometric.data import Dataset

from pointcyto.data.metadata_base import MetaDataBase
from pointcyto.io.parse_by_ext import parse_by_ext


class PointCloud(Dataset):
    """
    Currently **DEPRECATED**
    """

    def __init__(
        self,
        metadata: MetaDataBase,
        transform=None,
        pre_transform=None,
        clear_processed: bool = False,
        loading_function=parse_by_ext,
        loading_function_arguments=None,
    ):
        """
        Generates an InMemoryPointCloud based on the files inside metadata.

        In this root-directory, if not present, "raw/" and "processed/" directories are created.
        In "raw/", the original data can but does not have to lie.
        In "processed/", the processed data is saved.

        Args:
            metadata:
                A MetaDataBase object containing all relevant information
            transform:
                Transformations for the pytorch_geometric Data object
            pre_transform:
                Transformations for the pytorch_geometric Data object
            loading_function:
                function used in process() to load the data.
                Should return a dictionary with the following keys
                - ``features``: A list with names of the features
                - ``point_data``: A torch.Tensor with the actual loaded data.
            loading_function_arguments:
                Arguments additionally supplied to function used in process() to load the data.
                The function call is:

                .. code-block::

                    parsed = self.loading_function(
                        path=os.path.join(self.metadata.orig_dir, metadata_X["raw_filenames"]),
                        y=metadata_X["y"],
                        **self.loading_function_args
                    )

        Returns:
            No return
        """
        warnings.warn(
            "PointCloud is deprecated at the moment but not needed for probably up to 1k samples and "
            + "25k points per samples depending on your RAM - use InMemoryPointCloud until this is not "
            + "deprecated anymore.",
            DeprecationWarning,
        )

        self.loading_function = loading_function
        self.loading_function_args: Dict[str, Any]
        if loading_function_arguments is None:
            self.loading_function_args = {}
        else:
            self.loading_function_args = loading_function_arguments

        self.metadata = metadata
        if clear_processed:
            try:
                shutil.rmtree(os.path.join(self.metadata.root, "processed"))
            except FileNotFoundError:  # then it was already deleted
                pass

        # process is called inside here, therefore everything which is used in process must be defined before.
        super(PointCloud, self).__init__(
            root=metadata.root, transform=transform, pre_transform=pre_transform
        )

        previous_pointcloud_metadata_path = os.path.join(
            self.processed_dir, "pointcloud_metadata.pickle"
        )
        if os.path.exists(previous_pointcloud_metadata_path):
            # if it doesnt exist it may not have been initialized anytime before.
            with open(previous_pointcloud_metadata_path, "rb") as f:
                self.metadata = pickle.load(f)

    # Overwrite the raw_dir property from DataSet because
    # origdir might be something different than the "root/raw/" -directory.
    @property
    def raw_dir(self):
        return self.metadata.orig_dir

    @property
    def raw_file_names(self):
        # A list of files in the raw_dir which needs to be found in order to skip the download.
        return self.metadata.raw_filenames

    @property
    def processed_file_names(self):
        # A list of files in the processed_dir which needs to be found in order to skip the processing.
        return self.metadata.processed_filenames

    # I will never download data as they are assumed to be local.
    def download(self):
        # Would download raw data into raw_dir.
        pass

    def process(self):
        # Processes raw data and saves it into the processed_dir.
        # Read data into huge `Data` list.
        sample_feature_names = []
        for metadata_N, metadata_X in enumerate(self.metadata):
            print(
                str(metadata_N + 1),
                "/",
                str(len(self.metadata)),
                "  ",
                os.path.join(self.metadata.orig_dir, metadata_X["raw_filenames"]),
            )
            # Read data from `raw_path`.
            # now the data are in the format of tuple:(Tensor, markernames)
            # locals()["myfunction"]()
            parsed = self.loading_function(
                path=os.path.join(self.metadata.orig_dir, metadata_X["raw_filenames"]),
                y=metadata_X["y"],
                **self.loading_function_args
            )
            sample_feature_names += [parsed["features"]]
            if sample_feature_names[0] != parsed["features"]:
                raise ValueError(
                    "Unequal feature names of \n"
                    + "".join([self.metadata.raw_filenames[metadata_N]])
                    + ":"
                    + "\n  featurenames: "
                    + ", ".join(parsed["features"])
                    + "\n  sample 0:     "
                    + ", ".join(sample_feature_names[0])
                )
            data = parsed["point_data"]

            if self.pre_filter is not None:
                data = self.pre_filter(data)

            if self.pre_transform is not None:
                data = self.pre_transform(data)

            current_proc_dir = os.path.realpath(
                os.path.join(self.processed_paths[metadata_N], os.pardir)
            )
            if not os.path.exists(current_proc_dir):
                os.makedirs(current_proc_dir)

            # Todo: If I want to use different savetypes, self.metadata.savetype is the thing to look at
            #       and here is the place to insert it. (Probably also in __getitem__() or get())
            torch.save(data, self.processed_paths[metadata_N])
        self.metadata.sample_feature_names = sample_feature_names
        with open(
            os.path.join(self.processed_dir, "pointcloud_metadata.pickle"), "wb"
        ) as f:
            pickle.dump(self.metadata, file=f)

    def select_output_of_meta(self):
        # Todo: This must be done!
        raise NotImplementedError

    def len(self):
        return len(self.raw_file_names)

    def get(self, idx):
        return torch.load(self.processed_paths[idx])

    def __getitem__(self, item) -> "PointCloud":
        data = super(PointCloud, self).__getitem__(item)
        if not isinstance(item, int):
            print("\n\n\nasldfkjh")
            print(data.metadata)
            data.metadata = data.metadata[item, :]
            print(data.metadata)
            print("\n\n\n")
        return data



from typing import Any, List

import torch
from torch_geometric.data import Data


def convert_class_id_names(
    classes_id_or_name: List[str],
    class_name_to_id: dict,
    out: str = "name",
    allow_missings: bool = False,
) -> List[Any]:
    """
    Convert from class names/ids to either name or id. Used in MetaData.

    Args:
        classes_id_or_name:
            A list of either class_ids or class_names.
        class_name_to_id:
            dictionary with class_names as keys and class_ids as values
        out:
            Either "name" or "id", then the return values are either the name or the id. Fallback is "id"
        allow_missings:
            If true, missing lookup values in class_name_to_id are allowed and the result will either contain
            "__NotFound__"  (out="name") or -1 (out="id")

    Returns:

        ``out='name'``:

            List[str]

        ``out='id'``:

            List[int]



    """
    names = list(class_name_to_id.keys())
    ids = list(class_name_to_id.values())
    if out == "name":
        converted = ["__NotFound__"] * len(classes_id_or_name)
        search_in = ids
        search_for = names
    else:
        converted = [-1] * len(classes_id_or_name)
        search_in = names
        search_for = ids
    for index, val in enumerate(classes_id_or_name):
        try:
            converted[index] = search_for[search_in.index(val)]
        except ValueError as err:
            if allow_missings:
                pass
            else:
                raise ValueError("Did you already insert " + out + "?") from err
    return converted


def parse_matrix_class(point_position_matrix: torch.Tensor, class_id: int):
    """

    Args:
        point_position_matrix:
            A torch tensor beeing the data matrix for one sample
        class_id:
            The class_id for that sample

    Returns:
        A :class:`torch_geometric.data.Data` object.

    Todo:
        For non-neighbor approaches, the edge_index is probably not necessary at all and just costs space.
    """
    point_position_matrix = point_position_matrix.to(torch.float)
    # class_id = torch.empty(point_position_matrix.shape[0], 1).fill_(class_id)
    torch_geom_data = Data(pos=point_position_matrix, y=class_id)

    # # Add an empty but self-connected adjacency matrix.
    # # Each point must point to itself.
    # torch_geom_data.edge_index, unnecessary_edge_weight = add_self_loops(
    #     edge_index=torch.empty((2, 0), dtype=torch.long),
    #     num_nodes=torch_geom_data.pos.shape[0])
    # Replaced this by just an completely empty adjacency matrix and generate the self loops always on the fly
    # to save space
    torch_geom_data.edge_index = torch.empty((2, 0), dtype=torch.long)
    return torch_geom_data



from typing import List, Tuple

import torch


def pt_reader(*args, **kwargs):
    read_data = torch.load(*args, **kwargs)
    return read_data[0], read_data[1]


def read_pt(path: str, *args, **kwargs) -> Tuple[torch.Tensor, List[str]]:
    """
    read_jay reads the jay file (saved by datatable usually)

    Returns:
         Tuple of point cloud data and the column names

    """
    values_torch, colnames_list = pt_reader(path, *args, **kwargs)
    return values_torch, colnames_list



import copy
import os
import re
import warnings
from typing import Dict, List, Tuple, Union

import matplotlib.colors
import numpy as np
import pandas as pd

from pointcyto.data.metadata_base import MetaDataBase
from pointcyto.io.utils import convert_class_id_names
from pointcyto.plot.colors import discrete_cmap


# This class is heavily based on torch_geometric.data.data
class MetaDataClassification(MetaDataBase):
    """
    A plain python object containing pheno information about a sample
    """

    def __init__(
        self,
        raw_filenames: List[str] = None,
        read_with: List[str] = "parse_by_ext",
        class_name: Union[dict, List[str], pd.DataFrame] = None,
        orig_dir: str = os.path.abspath("."),
        root: str = None,
        savetype: str = "torch",
        allow_class_extension: bool = True,
        sort_class_name_map: bool = True,
        ignore_unequal_feature_names: bool = False,
    ):
        """

        Args:
            raw_filenames (List[str]):
                The raw filenames as list, relative path to orig_dir
            read_with:
                A list how the respective file should be read. Default: "ext" to read from file extension.
                2020-02-25: possible could be "parse_fcs", "parse_csv" or "sql"
            class_name:
                Either a list of class names for each filename
                or a dict of list of class names for each filename.
                Is transformed internally to a pandas.Dataframe. If no dict is given, the column names are generated
                automatically.
            orig_dir:
                The directory where raw_filenames are inside
            root:
                The directory where the folders "raw/" and "processed/" should be inside.
                In "root/processed/", the processed filenames will be saved.

                If root is not given, root is defined as inside the orig_dir.
            savetype:
                How the processed files should be saved. Currently (2020-02-25) only implemented: "torch" and "csv".
                Future extension could be "sql" but then the conversion of raw_filenames -> processed_filenames
                must be handled extra. At the moment, each raw_filename is directly translated
                (by changing the extension) to a processed filename
            allow_class_extension:
                When appending a new sample, including phenodata - is it allowed to rebuild the present classes?
                If the sample has an unseen class for one of the phenodata:

                If ``False``:

                    Throws a ValueError.

                If ``True``:

                    self._build_pheno_column() is called
            sort_class_name_map:
                If True, during building the class_name_id_map, the names are sorted before giving an ID.
                See the test tests/datasets/test_metadata.py: test_multiple_metadata_same_classmap() for exemplary use.
        """
        self._class_name_id_map: Dict[Dict] = dict()
        self._class_colors: Dict[matplotlib.colors.LinearSegmentedColormap] = dict()
        self._classes_base_cmap: Dict[str] = dict()
        self.allow_class_extension = allow_class_extension
        # sort_class_name_map is needed before super().__init__ because of self._build_pheno_column()
        self.sort_class_name_map = sort_class_name_map

        super(MetaDataClassification, self).__init__(
            raw_filenames=raw_filenames,
            read_with=read_with,
            orig_dir=orig_dir,
            root=root,
            savetype=savetype,
            pheno=class_name,
            ignore_unequal_feature_names=ignore_unequal_feature_names,
        )

        if self.raw_filenames is not None:
            self._build_pheno_column("read_with")

        self._listlike_metadata = self._listlike_metadata.union({"class_name"})

    def add_pheno_single(
        self,
        new_pheno_column: Union[pd.DataFrame, List],
        col_name: str = None,
        replace_new_df_index: bool = True,
    ) -> None:
        """
        Adds a single new pheno column to the ``mymeta.pheno``. Make sure that if you supply a dataframe that it really
        is only one column, if you have more than one column use :meth:`add_pheno`.

        Args:
            new_pheno_column:
                A dataframe or list of elements.
            col_name:
                How should the new column be named? You always have to give this.
            replace_new_df_index:
                If you gave a dataframe in ``new_pheno_column``, should the existing index be removed?
                Examples where this is relevant are in
                :func:`tests.datasets.test_metadata.test_add_pheno_single_missing_index` and
                :func:`tests.datasets.test_metadata.test_add_pheno_single_switched_index`.

        Returns:

        """
        if not isinstance(new_pheno_column, pd.DataFrame):
            new_pheno_column = pd.DataFrame({col_name: new_pheno_column})

        if self._pheno is not None:
            if not new_pheno_column.index.equals(self._pheno.index):
                warnmessage_part1 = "new_pheno_column.index is NOT equal to the current self.pheno.index **"
                if replace_new_df_index:
                    new_pheno_column.index = self._pheno.index
                    # warnings.warn(
                    #     warnmessage_part1 +
                    #     '\n  ** but I was allowed to replace the index (what I did now with self.pheno.index)')
                else:
                    warnings.warn(
                        warnmessage_part1
                        + "\n  ** your given new_pheno_column is reordered and/or can contain Nan!"
                    )

        if self._pheno is None:
            self._pheno = new_pheno_column
        else:
            if len(new_pheno_column) != len(self):
                raise ValueError("New pheno must be equal length as the raw_filenames")
            if col_name is None:
                if isinstance(new_pheno_column, pd.DataFrame):
                    col_name = new_pheno_column.columns[0]
                if col_name is None:
                    col_name = "pheno_" + str(len(self._pheno.columns))
            elif col_name == 0:
                col_name = "pheno_" + str(len(self._pheno.columns))

            self._pheno[col_name] = new_pheno_column
            self._build_pheno_column(col_name)

    def _is_pheno_column_valid(self, column_name: str):
        saved_selection = self.selection
        self.selection = column_name
        if not all(
            [
                value in list(self.class_name_id_map.keys())
                for value in self.class_name
                if value is not np.nan
            ]
        ):
            retval = False
        else:
            retval = True
        self.selection = saved_selection
        return retval

    def _build_pheno_column(self, column_name: str):
        """
        If a phenodata column is selected based on ``self.selection``, the following values are directly related to it:

        :meth:`class_name_id_map`
            A dictionary relating the class name to the class id

        :meth:`class_colors`
            Which class relates to which color (if you want to plot anytime this can then be consistent)

        :meth:`classes_base_cmap`
            The colormap for the selection which was the basis for class_colors

        These values come from the respective private (``__``) propierties. In constrast to the public property, the
        privates are each a dictionary of the following form:

        ::

            {
                pheno_column_X: values,
                pheno_column_Y: values,
            }

        To create these maps and colors ``__build_pheno_column`` is applied. Do not use that in your own analysis,
        instead if you really need it use :meth:`update_selection`.

        Args:
            column_name:
                Which of the phenodata should be rebuild.

        Returns:

        """
        saved_selection = self.selection
        self.selection = column_name
        # If an empty dict or None is supplied, the class_name_id_map is inferred from the data.
        self.class_name_id_map = None
        self.classes_base_cmap = "jet"
        self.class_colors = None  # If None is supplied, class_colors are inferred.
        if saved_selection is not None:
            self.selection = saved_selection

    def update_selection(self, sort_class_name_map: bool = True) -> None:
        """
        Usually you will not need this as the color and class/id maps are generated when adding phenodata. However if
        you need to update the class/id/colors for a specific selection you can call this and the class/id/colors data
        will be rebuilt.

        This might be usefull if you restrict your previously generated MetaDataClassification and therefore
        "loose" a unique value from all levels of the phenodata column.

        Always the current ``selection`` will be updated.

        Returns:
            None

        Example:
            :func:`tests.datasets.test_InMemoryPointCloud.test_InMemoryPointCloud_change_metadata`

        """
        self.sort_class_name_map = sort_class_name_map
        self._build_pheno_column(self.selection)

    def _del_pheno_column(self, column: Union[str, int]) -> None:
        """
        Only removing a column of self.pheno is not enough to really purge the column as in ``classes_base_cmap``,
        ``class_colors`` and ``class_name_id_map`` still the values exist. This function completely removes those and
        the column in the phenodata.

        Args:
            column:
                Column name or index

        Returns:
            None
        """
        del self._classes_base_cmap[column]
        del self._class_colors[column]
        del self._class_name_id_map[column]
        super(MetaDataClassification, self)._del_pheno_column(column)

    @property
    def class_name(self) -> List[str]:
        """
        Return the currently selected class names as a list, depending on ``self.selection``

        Returns:
            Usually List[str] but in principle it can also be List[Any], respective to raw_filenames.
        """
        return self.pheno[self.selection].tolist()

    @property
    def y(self):
        """
        Return the currently selected class IDs as a list, depending on ``self.selection``
        and ``self.class_name_id_map``

        Returns:
            List[int] according to the conversion accessible over ``self.class_name_id_map``
        """
        return convert_class_id_names(
            classes_id_or_name=self.pheno[self.selection],
            class_name_to_id=self._class_name_id_map[self.selection],
            out="id",
        )

    @property
    def class_name_id_map(self):
        """
        Returns:
            The current ``class_name_id_map``, a dictionary with the class names as keys and the class ids as values.

        Setter:

            Generated the class_name_id_map for the current selection.

            Args:
                name_id_map:
                    Default is None, then the class name and id map is extracted from the data.
                    If you supply this map directly:

                    .. code-block::

                        {
                            class_name_1: class_id_1,
                            class_name_2: class_id_2,
                            ...: ...,
                            class_name_k: class_id_k,
                        }

                    it is checked that all existing values in the currently selected pheno column (``self.selection``)
                    are inside this map.

                    If ``None`` is given (thus the class_name_id_map) is inferred from the data
                    the function additionally depends on if ``self.sort_class_name_map``.

                    If ``True`` (default):

                        The class names are sorted before assigning class ids

                    If ``False``:

                        The class names are NOT sorted before assigning class ids, therefore depending on
                        your input data if you split it for example into train and test that for train
                        ``classA = 0, classB = 1`` and for test  ``classA = 1, classB = 0``


        """
        return self._class_name_id_map[self.selection]

    @class_name_id_map.setter
    def class_name_id_map(self, name_id_map: dict = None):
        temp_class_name_id_map: Dict = dict()
        class_number = 0
        all_classes_set = list(
            dict.fromkeys(self.class_name)
        )  # unique values preserving order
        if self.sort_class_name_map:
            try:
                all_classes_set = sorted(all_classes_set)
            except TypeError as e:
                # THen was not able To compare values, returning UNSORTED + a warning
                warnings.warn(
                    "Did NOT sort the values in class_name_id_map because: \n" + str(e)
                )

        for class_x in all_classes_set:
            if class_x not in temp_class_name_id_map.keys():
                temp_class_name_id_map.update({class_x: class_number})
                class_number += 1
        if name_id_map is None or len(name_id_map) == 0:
            new_class_name_id_map = temp_class_name_id_map
        else:
            new_class_name_id_map = name_id_map
            missing_keys = []
            for key in temp_class_name_id_map:
                if key not in new_class_name_id_map:
                    missing_keys.append(key)
            if len(missing_keys) > 0:
                raise ValueError(
                    "Missing key(s) in supplied class_to_idx: ", missing_keys
                )

        if self.selection in self._class_name_id_map.keys():
            self._class_name_id_map[self.selection] = new_class_name_id_map
        else:
            self._class_name_id_map.update({self.selection: new_class_name_id_map})

    @property
    def class_colors(self) -> matplotlib.colors.Colormap:
        """
        Returns:
            The matplotlib.colormap for the current selection.
        """
        return self._class_colors[self.selection]

    @class_colors.setter
    def class_colors(self, cmap=None):
        if cmap is None:
            new_colormap = discrete_cmap(
                n=len(self.class_name_id_map), base_cmap=self.classes_base_cmap
            )
        else:
            new_colormap = cmap
        if self.selection in self._class_colors.keys():
            self._class_colors[self.selection] = new_colormap
        else:
            self._class_colors.update({self.selection: new_colormap})

    @property
    def classes_base_cmap(self):
        return self._classes_base_cmap[self.selection]

    @classes_base_cmap.setter
    def classes_base_cmap(self, base_cmap: str = "jet"):
        if self.selection in self._class_colors.keys():
            self._classes_base_cmap[self.selection] = base_cmap
        else:
            self._classes_base_cmap.update({self.selection: base_cmap})

    def class_id_color_dict(self) -> Dict[int, Tuple[float, float, float, float]]:
        """
        ``self.class_colors()`` returns a colormap which is of ``len(self.class_name_id_map)``. The color-coding is then
        1:1, so class_name_id become directly colors.
        class_id_color_dict is an actual dictionary making this explicit.

        .. code-block::

            {class_id: class_color}

        Returns:
            A dictionary:

            .. code-block::

                {class_id: class_color}

            where class_color is the result of self.class_colors(class_id),
            so a Tuple of RGBA values (see matplotlib.colors.Colormap)
        """
        return {id: self.class_colors(id) for id in self.class_name_id_map.values()}

    def class_name_color_dict(self) -> Dict[str, Tuple[float, float, float, float]]:
        """
        ``self.class_colors()`` returns a colormap which is of ``len(self.class_name_id_map)``. The color-coding is then
        1:1, so class_name_id become directly colors.
        class_name_color_dict is a actual dictionary:
        .. code-block::

            {'class_name': class_color}


        Returns:
            A dictionary:
            .. code-block::

                {class_name: class_color}

            where class_color is the result of self.class_colors(class_id) and class_id is gotten from class_id_name_map
            for the class_id. So a Tuple of RGBA values (see matplotlib.colors.Colormap)
        """
        class_id_name_map = {
            value: key for key, value in self.class_name_id_map.items()
        }
        return {
            class_id_name_map[id]: color
            for id, color in self.class_id_color_dict().items()
        }

    def append(self, new_listlike_dict: dict) -> None:
        """
        You can append new samples to the MetaDataClassification object using append.

        Args:
            new_listlike_dict:

        Returns:
            None

        Examples:
            See
            - :meth:`tests.datasets.test_metadata.test_metadata_append_only_filename`
            - :meth:`tests.datasets.test_metadata.test_metadata_append_with_sample_feature_names`
            - :meth:`tests.datasets.test_metadata.test_metadata_append_with_pheno`

        """
        super(MetaDataClassification, self).append(new_listlike_dict)
        try:
            tmp = pd.concat(
                [self.pheno, new_listlike_dict["pheno"]], sort=False, ignore_index=True
            )
            self._pheno = tmp
            for pheno_col in self._pheno.columns.values[1:]:
                if self.allow_class_extension:
                    self._build_pheno_column(pheno_col)
                elif not self._is_pheno_column_valid(pheno_col):
                    raise ValueError(
                        "Unknown class ",
                        new_listlike_dict["pheno"][pheno_col],
                        " in column ",
                        pheno_col,
                    )
        except KeyError:
            # Add a row of NA as last row to _pheno
            self._pheno.loc[self._pheno.shape[0]] = pd.NA

    def concat(
        self,
        other: "MetaDataClassification",
        inplace=False,
        ignore_identical_filenames: bool = False,
    ) -> "MetaDataClassification":
        # ignore_identical_filenames:
        #   Might be useful in some test cases? But usually not.
        if not inplace:
            self = copy.deepcopy(self)

        if not isinstance(other, MetaDataClassification):
            raise ValueError("Supply another MetaDataClassification object")
        if self.orig_dir != other.orig_dir:
            raise ValueError(
                "Can only concat MetaDataClassification with the same orig_dir "
                + "because raw_filenames are relative to that."
            )
        if not ignore_identical_filenames and any(
            [
                os.path.join(self.orig_dir, self_names)
                == os.path.join(other.orig_dir, other_names)
                for self_names, other_names in zip(
                    self.raw_filenames, other.raw_filenames
                )
            ]
        ):
            raise ValueError("Some (self.orig_dir, self_names) filenames are identical")
        self._raw_filenames += other.raw_filenames
        self._processed_filenames += other.processed_filenames
        self._sample_feature_names += other._sample_feature_names

        tmp = pd.concat([self.pheno, other.pheno], sort=False, ignore_index=True)
        self._pheno = tmp
        selection_saved = self.selection
        for pheno_col in self._pheno.columns.values[1:]:
            self.selection = pheno_col
            self.update_selection(sort_class_name_map=True)
        self.selection = selection_saved

        if not inplace:
            return self

    def __eq__(self, other):
        all_keys_present = all([key_x in other.keys for key_x in self.keys])
        if not all_keys_present:
            return False
        retval = True
        for key_x in self.keys:
            if key_x == "_class_colors":
                # class_colors is a matplotlib.colors.LinearSegmentedColormap object
                # which lives in a memory-place. This memory-place is compared
                # (and never equal) Thus, class_colors is just ignored during comparison.
                pass
            elif isinstance(getattr(self, key_x, None), pd.DataFrame):
                retval = getattr(self, key_x, None).equals(getattr(other, key_x, None))
            elif getattr(self, key_x, None) != getattr(other, key_x, None):
                retval = False
        return retval

    def __repr__(self):
        base_repr = super(MetaDataClassification, self).__repr__()

        if len(self) == 0:
            selected_class_levels = "None"
        else:
            selected_class_levels = str(list(self.class_name_id_map.keys()))
        # inject _after_ "selection" line
        base_repr = re.sub(
            pattern=r"(\nselection:.*\n)",
            repl=r"\1___AAAAAAAAAAAAAAAAAAAAAAAA___PLACEHOLDER___\n",
            string=base_repr,
        )
        return re.sub(
            "___AAAAAAAAAAAAAAAAAAAAAAAA___PLACEHOLDER___",
            "selected class levels: {}".format(selected_class_levels),
            base_repr,
        )



import os
import shutil
from unittest import TestCase

import torch

from pointcyto.data.InMemoryPointCloud import InMemoryPointCloud
from pointcyto.io.meta_read_foldering import gen_foldering_meta
from pointcyto.testutils.helpers import find_dirname_above

THIS_DIR = os.path.dirname(os.path.abspath(__file__))
# Now build the base-test directory (necessary if you want to use tests/testdata)
TESTS_DIR = find_dirname_above(THIS_DIR)

pointcloud_toy_dataset = os.path.join(
    "testdata", "flowcytometry", "PointClouds_toy_dataset"
)
tcell_toy_dataset = os.path.join("testdata", "flowcytometry", "Tcell_foldering")
empty_sample_dataset = os.path.join("testdata", "flowcytometry", "empty_sample")


class TestInMemoryPointCloud(TestCase):
    def test_InMemoryPointCloud(self):
        orig_data_dir = os.path.join(TESTS_DIR, pointcloud_toy_dataset, "raw")
        new_data_dir = os.path.join(TESTS_DIR, pointcloud_toy_dataset, "removeme_raw")
        shutil.copytree(src=orig_data_dir, dst=new_data_dir, dirs_exist_ok=True)
        # generate the metadata
        mymeta = gen_foldering_meta(new_data_dir)
        # make sure that this dir was previously empty
        shutil.rmtree(mymeta.root, ignore_errors=True)

        # Process the data the first time
        a = InMemoryPointCloud(metadata=mymeta)
        assert len(a) == 6

        # Moving the rootdir should be fine, just the data is preprocessed again
        mymeta.root = os.path.join(new_data_dir, os.pardir, "autoroot_second")
        # make sure that this dir was previously empty
        shutil.rmtree(mymeta.root, ignore_errors=True)

        # Now the following should break because neither the root dir exists, nor the original data
        b = InMemoryPointCloud(
            metadata=mymeta
        )  # that must fail now because the data are missing? NO?
        assert len(b) == 6

        # Now forcefully remove the original data
        # Which is mimiking the behaviour of going to a different server!
        shutil.rmtree(new_data_dir)

        # should be processed automatically
        # that should be fine because the data was already processed
        #   It searches under os.path.join(mymeta.root, 'processed')
        b = InMemoryPointCloud(metadata=mymeta)
        assert len(b) == 6

        # print(mymeta.root)
        # # /home/gugl/programming/ccc/tests/testdata/flowcytometry/PointClouds_toy_dataset/autoroot_removeme_raw
        mymeta.root = os.path.join(new_data_dir, os.pardir, "autoroot_removeme_raw")
        # that should ALSO be fine because the data was already processed, I just want to check
        # if changing the root breaks anything unintentionally
        #   It searches under os.path.join(mymeta.root, 'processed')
        b = InMemoryPointCloud(metadata=mymeta)
        assert len(b) == 6

        mymeta.root = os.path.join(new_data_dir, os.pardir, "NEW_UNSEEN_autoroot")
        # Now the following should break because neither the root dir exists, nor the original data
        with self.assertRaises(FileNotFoundError):
            b = InMemoryPointCloud(metadata=mymeta)

    def test_reload_processed_meta(self):
        orig_data_dir = os.path.join(TESTS_DIR, pointcloud_toy_dataset, "raw")
        new_data_dir = os.path.join(TESTS_DIR, pointcloud_toy_dataset, "removeme_raw")
        shutil.copytree(src=orig_data_dir, dst=new_data_dir, dirs_exist_ok=True)
        # generate the metadata
        mymeta = gen_foldering_meta(new_data_dir)
        # make sure that this dir was previously empty
        shutil.rmtree(mymeta.root, ignore_errors=True)

        # Process the data the first time
        a = InMemoryPointCloud(metadata=mymeta)
        assert len(a) == 6

        import pickle

        with open(
            os.path.join(a.processed_dir, "pointcloud_metadata.pickle"), "rb"
        ) as f:
            x = pickle.load(f)

        a_reloaded = InMemoryPointCloud(metadata=x)
        assert a_reloaded.metadata == a.metadata
        assert torch.equal(a_reloaded.pos, a.pos)

        # now move the complete processed folder
        new_autoroot = os.path.join(
            TESTS_DIR, pointcloud_toy_dataset, "removeme_new_autoroot"
        )
        shutil.rmtree(new_autoroot, ignore_errors=True)
        shutil.copytree(src=a.root, dst=new_autoroot)

        # remove all processed folders
        shutil.rmtree(a.root, ignore_errors=True)
        # remove the original data such that RE-processing would throw an error
        shutil.rmtree(new_data_dir, ignore_errors=True)

        with self.assertRaises(FileNotFoundError):
            a_moved_reloaded = InMemoryPointCloud(metadata=x)

        # But if we change the metadata root to the newly copied autoroot it should work again
        x.root = new_autoroot
        a_moved_reloaded = InMemoryPointCloud(metadata=x)

        # test it with new_autoroot as string
        a_moved_reloaded = InMemoryPointCloud(metadata=new_autoroot)
        print(a_moved_reloaded)

    def test_reload_processed_meta_pre_transform_warning(self):
        orig_data_dir = os.path.join(TESTS_DIR, pointcloud_toy_dataset, "raw")
        # generate the metadata
        mymeta = gen_foldering_meta(orig_data_dir)
        # make sure that this dir was previously empty
        shutil.rmtree(mymeta.root, ignore_errors=True)

        from torch_geometric.transforms.fixed_points import FixedPoints

        # Process the data the first time
        a = InMemoryPointCloud(metadata=mymeta, pre_transform=FixedPoints(10))
        # now reloading, should not give a warning
        import warnings

        # b = InMemoryPointCloud(metadata=a.root)
        with warnings.catch_warnings():
            warnings.simplefilter("error")
            b = InMemoryPointCloud(metadata=a.root)
            print(b)



import os

# from pathlib import Path, PurePath
import pathlib
import re
import sys
from typing import Dict, List, Tuple, Union

import pandas as pd

from pointcyto.data.metadata_base import MetaDataBase
from pointcyto.data.metadata_classification import MetaDataClassification
from pointcyto.io.utils import convert_class_id_names


# Taken from https://github.com/pytorch/vision/blob/master/torchvision/datasets/folder.py
def has_file_allowed_extension(filename: str, extensions: Tuple[str]) -> bool:
    """
    Checks if a file is an allowed extension.

    Args:
        filename (string):
            path to a file
        extensions (tuple of strings):
            extensions to consider (lowercase)

    Returns:
        bool: True if the filename ends with one of given extensions
    """
    return filename.lower().endswith(extensions)


def make_dataset(
    basedir,
    class_to_idx: dict,
    extensions: Tuple[str, ...] = (".csv", ".fcs"),
    is_valid_file=None,
) -> List[Tuple[str, int]]:
    """
    Traverse through all (direct) subfolders of basedir
    Each subfolder is the respective class for all files inside

    Args:
        basedir:
            The root directory which should be searched
        class_to_idx:
            Dictionary to enumerate the subfolders/classes. E.g. {classA: 0, classB: 1}
        extensions:
            File extensions which should be seen as valid, checked by has_file_allowed_extension
        is_valid_file:
            Alternatively give a function which says if the file is valid

    Returns:
         List of tuples with (filepath, classID)
    """
    samples_path_label = []

    if not ((extensions is None) ^ (is_valid_file is None)):
        raise ValueError(
            "Both extensions and is_valid_file cannot be None or not None at the same time"
        )
    if extensions is not None:

        def is_valid_file(file_x):
            return has_file_allowed_extension(file_x, extensions)

    # Traverse through all subfolders of self.raw_dir
    # Each subfolder is the respective class for all files inside
    for target in sorted(class_to_idx.keys()):
        tmp_dir = os.path.join(basedir, target)
        if not os.path.isdir(tmp_dir):
            continue
        for root, _, fnames in sorted(os.walk(tmp_dir, followlinks=True)):
            for fname in sorted(fnames):
                path = os.path.join(root, fname)
                if is_valid_file(path):
                    item = (path, class_to_idx[target])
                    samples_path_label.append(item)
    return samples_path_label


def find_classes(whichdir: str) -> Tuple[List[str], Dict[str, int]]:
    """
    Finds the class folders in whichdir. Not recursive.

    Args:
        whichdir:
            The starting directory where all directories are listed.

    Returns:
        tuple: (classes, class_to_idx) where classes are relative to (dir), and class_to_idx is a dictionary.

    Ensures:
        No class is a subdirectory of another.
    """
    if sys.version_info >= (3, 5):
        # Faster and available in Python 3.5 and above
        classes = [d.name for d in os.scandir(whichdir) if d.is_dir()]
    else:
        classes = [
            d for d in os.listdir(whichdir) if os.path.isdir(os.path.join(whichdir, d))
        ]
    classes.sort()
    class_to_idx = {classes[i]: i for i in range(len(classes))}
    return classes, class_to_idx


def gen_foldering_meta(
    orig_dir: str, root: str = None, **kwargs
) -> MetaDataClassification:
    """
    Generate a MetaData object based on "foldered" data. Looks as following:

    orig_dir/
        A/
            file1
            file2
            file3
        B/
            file4
            file5

    Then the MetaData object will be of length 5 and have two classes: "A" and "B".

    Args:
        orig_dir:
            Data containing the raw data in folders. Each direct sub-folder of this directory is regarded as unique
            class name.
        root:
            The root dir might differ from orig_dir. If given, MetaData processed/ directory is created inside this
            root. If not given, MetaData processed/ directory is created in "orig_dir/../processed"
        **kwargs:
            Further arguments to MetaData

    Returns:
        A MetaData object
    """
    classes_classtoidx = find_classes(orig_dir)
    fullpath_class_metadata = make_dataset(orig_dir, classes_classtoidx[1])
    #  len(orig_dir)+1 to get rid of the \
    relative_paths_class = [
        (path[len(orig_dir) + 1 :], class_id)
        for path, class_id in fullpath_class_metadata
    ]
    relative_paths = [path for path, class_id in relative_paths_class]
    class_ids = [class_id for path, class_id in relative_paths_class]
    class_names = convert_class_id_names(class_ids, classes_classtoidx[1], out="name")
    return MetaDataClassification(
        raw_filenames=relative_paths,
        orig_dir=orig_dir,
        class_name=class_names,
        root=root,
        **kwargs
    )


def list_files_recursive(dirpath, full_names: bool = False) -> List[str]:
    """
    List all files inside the dirpath recursively

    Args:
        dirpath:
            Starting point for recursion
        full_names:
            should full paths be returned?

    Returns:
        List of file paths.
    """
    full_paths = [
        os.path.join(dp, f) for dp, dn, filenames in os.walk(dirpath) for f in filenames
    ]
    if full_names:
        retval = full_paths
    else:
        retval = [filename[len(dirpath) + 1 :] for filename in full_paths]
    return retval


def gen_csv_meta(
    orig_dir: str,
    path_phenodata: str,
    regex_include: str = r".fcs$",
    read_csv_kwargs: dict = None,
    subject_name_col: Union[int, str] = 0,
    regex_sub_subject_name_col: Tuple[str, str] = None,
    restrict_to_present_pheno: bool = False,
    verbose: bool = False,
    metadata_type=MetaDataClassification,
    **kwargs
) -> MetaDataBase:
    """
    Generate a MetaData object based on data plus phenodata

    orig_dir/
        file1
        file2
        A/
            file3
            file4
        pheno_file.csv

    Where in pheno_file.csv there must be atleast 2 columns (but can be more):
        column1, column2, column3
        file1, 1, A
        file2, 2, B
        A/file3, 0, A
        A/file4, 0, A
    The first column always holds the path to the files.
    Second( or more) columns hold classes.

    Then the MetaData object will be of length 4 and have two possible "selection"s: column2 and column3.
    Selection column2 has 3 classes: 0, 1, 2
    Selection column3 has 2 classes: "A", "B"

    Args:
        orig_dir:
            The directory where all files (possibly in multiple subfolders) are located
        path_phenodata:
            path to phenodata. The phenodata file
                - must have the first column as paths to the files, relative to orig_dir
                - must have all found files in origdir in the first column
                - should have atleast 2 columns (including the filepath-column)
                - can have multiple columns
                - can have more rows than files, only the matched to filenames are selected
        read_csv_kwargs:
            Arguments given (**read_csv_kwargs) to pandas.read_csv
            (Used in examples/02_gen_MetaData/06_Putzel_meta.py)
        regex_include:
            Include only files to the metadata matching this regex. Default: r'.fcs' --> all .fcs or .FCS files
            Ignore case because , "re.IGNORECASE"
        subject_name_col:
            Default 0: The column of path_phenodata containing the filenames of the files to load.
            Can be numeric (column number) or string (exact column name)
        restrict_to_present_pheno:
            If True, only the files are included where phenodata[subject_name_col] is present.
        **kwargs:
            Further arguments to MetaDataClassification()

    Returns:
        MetaData object
    """

    if not os.path.exists(path_phenodata):
        path_phenodata = os.path.join(orig_dir, path_phenodata)
        if not os.path.exists(path_phenodata):
            raise FileNotFoundError("Phenodata file not found")
    if read_csv_kwargs is not None:
        phenodata = pd.read_csv(path_phenodata, **read_csv_kwargs)
    else:
        phenodata = pd.read_csv(path_phenodata)
    # If string values, remove trailing and leading whitespace
    for column_x in phenodata.columns:
        if isinstance(phenodata[column_x][0], str):
            phenodata[column_x] = phenodata[column_x].str.strip()

    if isinstance(subject_name_col, int):
        subject_col = phenodata.columns.values[subject_name_col]
    elif isinstance(subject_name_col, str):
        subject_col = subject_name_col

    pheno_subject_paths: List[str] = []
    for x in phenodata[subject_col].tolist():
        if regex_sub_subject_name_col is not None:
            x = re.sub(regex_sub_subject_name_col[0], regex_sub_subject_name_col[1], x)
        from_posix_path = pathlib.Path(pathlib.PurePosixPath(x))
        from_nt_path = pathlib.Path(pathlib.PureWindowsPath(x))
        if os.path.exists(os.path.join(orig_dir, from_posix_path)):
            found_path = str(from_posix_path)
        elif os.path.exists(os.path.join(orig_dir, from_nt_path)):
            found_path = str(from_nt_path)
        else:
            found_path = x + "___not_found_in_orig_dir"
        pheno_subject_paths += [found_path]
    phenodata[subject_col] = pheno_subject_paths
    phenodata.set_index(subject_col, inplace=True)
    if verbose:
        print("phenodata", phenodata)

    # find all files inside origdir
    all_files = list_files_recursive(orig_dir, full_names=False)
    if verbose:
        print("all_files", all_files)

    # include only matching regex
    included_files = [
        filename
        for filename in all_files
        if bool(re.search(regex_include, filename, re.IGNORECASE))
    ]
    if verbose:
        print("included_files", included_files)

    # sort all_files because windows and linux seem to sort differently, then the test assertion goes wrong.
    included_files.sort()
    files_not_in_pheno = [
        filex
        for filex in included_files
        if pathlib.Path(filex).__str__() not in pheno_subject_paths
    ]
    if verbose:
        print("files_not_in_pheno", files_not_in_pheno)

    if restrict_to_present_pheno:
        # if restrict to present pheno: restrict the included files to those present in the phenodata.
        included_files = [
            filex
            for filex in included_files
            if pathlib.Path(filex).__str__() in pheno_subject_paths
        ]
    else:
        if len(files_not_in_pheno) > 0:
            raise ValueError(
                "The following file indexes are missing in the phenodata: \n"
                + ", ".join(files_not_in_pheno)
            )
    if verbose:
        print("included_files", included_files)

    included_files_pathed = [pathlib.Path(filex).__str__() for filex in included_files]
    if verbose:
        print("included_files_pathed", included_files_pathed)

    # match the phenodata with the read in files
    matched_pheno = phenodata.loc[included_files_pathed, :]
    if verbose:
        print("matched_pheno", matched_pheno)

    if any(phenodata.index.duplicated()):
        raise ValueError(
            "You have duplicated phenodata rownames (even after subsetting to included files!)\n"
            + "--> Check your csv file. "
        )
    if len(included_files) == 0:
        raise ValueError("Empty included_files.")

    matched_pheno.reset_index(drop=False, inplace=True)
    # matched_pheno.reset_index(drop=True, inplace=True)
    if verbose:
        print("matched_pheno", matched_pheno)

    return metadata_type(
        raw_filenames=included_files,
        orig_dir=orig_dir,
        class_name=matched_pheno,
        **kwargs
    )


def gen_meta_filenames(
    orig_dir: str,
    colnames: List[str] = None,
    sep: str = r"_",
    regex_include: str = r"^",
    verbose: bool = False,
    metadata_type: MetaDataBase = MetaDataClassification,
    **kwargs
) -> MetaDataBase:
    """
    Generate a MetaData object based on the data's filenames, delimited through the "sep" separator.

    orig_dir/
        file1_1_A
        file2_0_A
        file3_2_B
        file4_0_B
        file5_1_B

    Then the MetaData object will be of length 5 and have two possible "selection"s: column2 and column3.
    Selection column2 has 3 classes: 0, 1, 2
    Selection column3 has 2 classes: "A", "B"

    So the interpretation of the filename works as following where <sep> is replaced be the value of sep:

        samplename<sep>column2<sep>column3

    Args:
        orig_dir:
            Data containing the raw data in folders. Each direct sub-folder of this directory is regarded as unique
            class name.
        colnames:
            If None, the colnames are set as following:
                ['sample', *['column' + str(i) for i in range(1, len(filenames_split_df.columns))]]
            resulting in something like
                ['sample', 'column1', 'column2', ..., 'columnK']
            By setting colnames you can replace these names.
        sep:
            Separator between the "columns" inside the filenames
        regex_include:
            Include only files to the metadata matching this regex. Default: r'*' --> all files
        verbose:
            Verbose output
        **kwargs:
            Further arguments to MetaData

    Returns:
        A MetaData object
    """
    # find all files inside origdir
    all_files = list_files_recursive(orig_dir, full_names=False)
    all_files.sort()  # consistency between windows and linux
    if verbose:
        print(all_files)

    # include only matching regex
    included_files = [
        filename for filename in all_files if bool(re.search(regex_include, filename))
    ]
    if verbose:
        print(included_files)

    # remove the extensions, as those are certainly no "column"
    # os.path.splitext: Split a path in root [0] and extension [1].
    matched_files_no_extension = [
        os.path.splitext(filename)[0] for filename in included_files
    ]
    if verbose:
        print(matched_files_no_extension)

    # split the "columns" by the separator
    filenames_split = [
        re.split(sep, filename) for filename in matched_files_no_extension
    ]
    if verbose:
        print(filenames_split)

    # make a dataframe and set appropriate names
    filenames_split_df = pd.DataFrame(filenames_split)
    if colnames is not None:
        if len(colnames) < len(filenames_split_df.columns):
            raise ValueError(
                "You supplied less colnames ("
                + str(len(colnames))
                + ") than there are in the filenames_split dataframe ("
                + str(len(filenames_split_df.columns))
                + ")"
            )
        elif len(colnames) > len(filenames_split_df.columns):
            raise ValueError(
                "You supplied more colnames ("
                + str(len(colnames))
                + ") than there are in the filenames_split dataframe ("
                + str(len(filenames_split_df.columns))
                + ")"
            )
        filenames_split_df.columns = colnames

    else:
        filenames_split_df.columns = [
            "sample",
            *["column" + str(i) for i in range(1, len(filenames_split_df.columns))],
        ]
    if verbose:
        print(filenames_split_df)

    return metadata_type(
        raw_filenames=included_files,
        orig_dir=orig_dir,
        class_name=filenames_split_df,
        **kwargs
    )



import os
import shutil
from unittest import TestCase

import torch
from torch_geometric.transforms.fixed_points import FixedPoints

from pointcyto.data.datamodule import PCDataModule
from pointcyto.data.PointCloud import PointCloud
from pointcyto.io.meta_read_foldering import gen_foldering_meta
from pointcyto.io.pickle_open_dump import pickle_open_dump
from pointcyto.testutils.helpers import find_dirname_above_currentfile

# Now build the base-test directory (necessary if you want to use tests/testdata)
TESTS_DIR = find_dirname_above_currentfile()


pointcloud_toy_dataset = os.path.join(
    "testdata", "flowcytometry", "PointClouds_toy_dataset"
)
tcell_toy_dataset = os.path.join("testdata", "flowcytometry", "Tcell_foldering")


class TestPCDataModule(TestCase):
    def test_PointCloud(self):
        mymeta = gen_foldering_meta(
            os.path.join(TESTS_DIR, pointcloud_toy_dataset, "raw")
        )
        a = PointCloud(metadata=mymeta, clear_processed=True)
        a = PointCloud(metadata=mymeta)
        shutil.rmtree(os.path.join(TESTS_DIR, pointcloud_toy_dataset, "autoroot_raw"))
        a = PointCloud(metadata=mymeta)
        repr(a)

    def test_DataModule(self):
        mymeta = gen_foldering_meta(
            os.path.join(TESTS_DIR, pointcloud_toy_dataset, "raw")
        )
        with self.assertRaises(ValueError):
            # Cannot create a PCDataModule with only one dataset. Use (InMemory)PointCloud
            dm = PCDataModule(meta_train=mymeta)
        dm = PCDataModule(meta_train={"dataA": mymeta})
        print(dm)

    def test_repr(self):
        mymeta = gen_foldering_meta(
            os.path.join(TESTS_DIR, pointcloud_toy_dataset, "raw")
        )
        with self.assertRaises(ValueError):
            dm = PCDataModule(meta_train=mymeta)
        dm = PCDataModule(meta_train={"dataA": mymeta})
        print(dm)

    def test_meta_inputs(self):
        mymeta = gen_foldering_meta(
            os.path.join(TESTS_DIR, pointcloud_toy_dataset, "raw")
        )

        pickle_open_dump(mymeta, "mymeta.pickle")
        with self.assertRaises(ValueError):
            dm = PCDataModule(meta_train=mymeta)

        dm = PCDataModule(
            meta_train={
                "dataA": mymeta,
                "dataB": "mymeta.pickle",
                "dataC": {"metadata": mymeta},
            }
        )
        assert dm.meta_conf_xxx["train"]["dataA"] == dm.meta_conf_xxx["train"]["dataB"]
        assert dm.meta_conf_xxx["train"]["dataA"] == dm.meta_conf_xxx["train"]["dataC"]

    def test_meta_missingParts(self):
        mymeta = gen_foldering_meta(
            os.path.join(TESTS_DIR, pointcloud_toy_dataset, "raw")
        )
        with self.assertRaises(ValueError):
            # Missing dataB in meta_test
            dm = PCDataModule(
                meta_train={
                    "dataA": mymeta,
                    "dataB": mymeta,
                },
                meta_test={"dataA": mymeta},
            )

        dm = PCDataModule(
            meta_train={
                "dataA": mymeta,
                "dataB": mymeta,
            },
            meta_test={
                "dataA": mymeta,
                "dataB": mymeta,
            },
        )
        assert dm.meta_conf_xxx["train"]["dataA"] == dm.meta_conf_xxx["train"]["dataB"]
        assert dm.meta_conf_xxx["train"]["dataA"] == dm.meta_conf_xxx["test"]["dataA"]
        assert dm.meta_conf_xxx["train"]["dataA"] == dm.meta_conf_xxx["test"]["dataB"]

    def test_prepare(self):
        mymeta = gen_foldering_meta(
            os.path.join(TESTS_DIR, pointcloud_toy_dataset, "raw")
        )

        dm = PCDataModule(
            meta_train={
                "dataA": mymeta,
                "dataB": mymeta,
            },
            meta_test={
                "dataA": mymeta,
                "dataB": mymeta,
            },
        )
        assert dm.must_prepare
        assert list(dm.meta_conf_xxx["train"]["dataA"].keys()) == ["metadata"]
        dm.prepare_data()
        assert list(dm.meta_conf_xxx["train"]["dataA"].keys()) == [
            "metadata",
            "clear_processed",
        ]
        assert not dm.must_prepare  # gets changed by .prepare_data()
        assert not dm.meta_conf_xxx["train"]["dataA"]["clear_processed"]
        assert not dm.meta_conf_xxx["train"]["dataB"]["clear_processed"]
        assert not dm.meta_conf_xxx["test"]["dataA"]["clear_processed"]
        assert not dm.meta_conf_xxx["test"]["dataB"]["clear_processed"]

    def test_setup(self):
        mymeta = gen_foldering_meta(
            os.path.join(TESTS_DIR, pointcloud_toy_dataset, "raw")
        )

        dm = PCDataModule(
            meta_train={
                "dataA": mymeta,
                "dataB": mymeta,
            },
            meta_test={
                "dataA": mymeta,
                "dataB": mymeta,
            },
        )
        with self.assertRaises(ValueError):
            dm.setup("train")

        dm.prepare_data()
        dm.setup("train")

        print(dm.pc_xxx)
        assert torch.allclose(
            dm.pc_xxx["train"]["dataA"].pos, dm.pc_xxx["train"]["dataB"].pos
        )
        with self.assertRaises(KeyError):
            torch.allclose(
                dm.pc_xxx["test"]["dataA"].pos, dm.pc_xxx["test"]["dataB"].pos
            )

        dm.setup("test")
        assert torch.allclose(
            dm.pc_xxx["test"]["dataA"].pos, dm.pc_xxx["test"]["dataB"].pos
        )

    def test_clear_processed(self):
        # From this test, the printed output should be  4 x 2 times from InMemoryPointCloud.process()
        mymeta = gen_foldering_meta(
            os.path.join(TESTS_DIR, pointcloud_toy_dataset, "raw")
        )

        dm = PCDataModule(
            meta_train={
                "dataA": mymeta,
                "dataB": mymeta,
            },
            meta_test={
                "dataA": mymeta,
                "dataB": mymeta,
            },
            clear_processed=True,
        )
        dm.prepare_data()
        dm = PCDataModule(
            meta_train={
                "dataA": mymeta,
                "dataB": mymeta,
            },
            meta_test={
                "dataA": mymeta,
                "dataB": mymeta,
            },
            clear_processed=False,
        )
        dm.prepare_data()
        dm = PCDataModule(
            meta_train={
                "dataA": mymeta,
                "dataB": mymeta,
            },
            meta_test={
                "dataA": mymeta,
                "dataB": mymeta,
            },
            clear_processed=False,
        )
        dm.prepare_data(clear_processed=True)

    def test_setup_None(self):
        mymeta = gen_foldering_meta(
            os.path.join(TESTS_DIR, pointcloud_toy_dataset, "raw")
        )

        dm = PCDataModule(
            meta_train={
                "dataA": mymeta,
                "dataB": mymeta,
            },
            meta_test={
                "dataA": mymeta,
                "dataB": mymeta,
            },
            clear_processed=True,
        )
        dm.prepare_data()
        dm.setup()
        assert torch.allclose(
            dm.pc_xxx["train"]["dataA"].pos, dm.pc_xxx["train"]["dataB"].pos
        )
        assert torch.allclose(
            dm.pc_xxx["test"]["dataA"].pos, dm.pc_xxx["test"]["dataB"].pos
        )

    def test_check_Dataloader(self):
        mymeta = gen_foldering_meta(
            os.path.join(TESTS_DIR, pointcloud_toy_dataset, "raw")
        )

        dm = PCDataModule(
            meta_train={
                "dataA": mymeta,
                "dataB": mymeta,
            },
            meta_test={
                "dataA": mymeta,
                "dataB": mymeta,
            },
            clear_processed=True,
            batch_size=4,  # there are 6 samples, therefore now there should be 2 batches
        )
        dm.prepare_data()

        with self.assertRaises(KeyError):
            for loader in dm.train_dataloader():
                print(loader["dataA"])

        dm.setup("train")

        counter = 0
        # for loader in dm.train_dataloader():
        for batch, batch_idx, dataloader_idx in dm.train_dataloader():
            assert counter == batch_idx
            counter += 1
            assert isinstance(batch, dict)
            assert list(batch.keys()) == ["dataA", "dataB"]
            assert torch.allclose(batch["dataA"].pos, batch["dataB"].pos)
            assert dataloader_idx == 0  # always the same dataloader is used here
        assert counter == 2

    def test_check_pointcloud_args_datapart(self):
        mymeta = gen_foldering_meta(
            os.path.join(TESTS_DIR, pointcloud_toy_dataset, "raw")
        )
        with self.assertRaises(ValueError):
            dm = PCDataModule(
                meta_train={
                    "dataA": mymeta,
                },
                meta_val={
                    "dataA": mymeta,
                },
                clear_processed=True,
                batch_size=4,  # there are 6 samples, therefore now there should be 2 batches
                pointcloud_args_datapart={
                    "transform": FixedPoints(
                        num=100, replace=False, allow_duplicates=True
                    )
                },
            )
        dm = PCDataModule(
            meta_train={
                "dataA": mymeta,
            },
            meta_val={
                "dataA": mymeta,
            },
            clear_processed=True,
            batch_size=4,  # there are 6 samples, therefore now there should be 2 batches
            pointcloud_args_datapart={
                "dataA": {
                    "transform": FixedPoints(
                        num=100, replace=False, allow_duplicates=True
                    )
                }
            },
        )
        print(dm)

    def tearDown(self) -> None:
        try:
            shutil.rmtree(os.path.join(TESTS_DIR, pointcloud_toy_dataset, "processed"))
            pass
        except FileNotFoundError:
            pass
        try:
            shutil.rmtree(
                os.path.join(TESTS_DIR, pointcloud_toy_dataset, "raw", "processed")
            )
            shutil.rmtree(os.path.join(TESTS_DIR, pointcloud_toy_dataset, "raw", "raw"))
            pass
        except FileNotFoundError:
            pass



import os
import shutil
from unittest import TestCase

import torch
from torch_geometric.loader import DataLoader as tg_DataLoader
from torch_geometric.transforms.fixed_points import FixedPoints

from pointcyto.data.PointCloud import PointCloud
from pointcyto.io.meta_read_foldering import gen_foldering_meta
from pointcyto.testutils.helpers import find_dirname_above

THIS_DIR = os.path.dirname(os.path.abspath(__file__))
# Now build the base-test directory (necessary if you want to use tests/testdata)
TESTS_DIR = find_dirname_above(THIS_DIR)

pointcloud_toy_dataset = os.path.join(
    "testdata", "flowcytometry", "PointClouds_toy_dataset"
)
tcell_toy_dataset = os.path.join("testdata", "flowcytometry", "Tcell_foldering")
empty_sample_dataset = os.path.join("testdata", "flowcytometry", "empty_sample")


class TestPointCloud(TestCase):
    def test_PointCloud(self):
        mymeta = gen_foldering_meta(
            os.path.join(TESTS_DIR, pointcloud_toy_dataset, "raw")
        )
        a = PointCloud(metadata=mymeta, clear_processed=True)
        a = PointCloud(metadata=mymeta)
        shutil.rmtree(os.path.join(TESTS_DIR, pointcloud_toy_dataset, "autoroot_raw"))
        a = PointCloud(metadata=mymeta)
        repr(a)

    def test_Pointcloud_item(self):
        mymeta = gen_foldering_meta(
            os.path.join(TESTS_DIR, pointcloud_toy_dataset, "raw")
        )
        a = PointCloud(metadata=mymeta, clear_processed=True)
        # When a slice is selected, the returned value is again an InMemoryPointCloud
        print(a[0:2])
        # If an int is selected, the returned value is a torch_geometric Data instance.
        print(a[0])

    def test_PointCloud_subsetting(self):
        mymeta = gen_foldering_meta(
            os.path.join(TESTS_DIR, pointcloud_toy_dataset, "raw")
        )
        mymeta_train = mymeta[0:2, :]
        mymeta_val = mymeta[2:, :]
        a_train = PointCloud(metadata=mymeta_train, clear_processed=True)
        a_val = PointCloud(metadata=mymeta_val, clear_processed=True)
        assert len(a_train) == 2
        assert len(a_val) == 4

        complete_pointcloud = PointCloud(metadata=mymeta, clear_processed=True)
        print(complete_pointcloud.metadata)
        pc_train = complete_pointcloud[0:2]
        pc_val = complete_pointcloud[2:]

        assert all(mymeta_train.pheno == pc_train.metadata.pheno)
        assert all(mymeta_val.pheno == pc_val.metadata.pheno)

        assert (
            mymeta_train.sample_feature_names == pc_train.metadata.sample_feature_names
        )
        assert mymeta_val.sample_feature_names == pc_val.metadata.sample_feature_names

    def test_reloadPointcloud(self):
        mymeta = gen_foldering_meta(
            os.path.join(TESTS_DIR, pointcloud_toy_dataset, "raw")
        )
        complete_pointcloud = PointCloud(metadata=mymeta, clear_processed=True)
        loaded_pointcloud = PointCloud(metadata=mymeta)
        assert complete_pointcloud.metadata == loaded_pointcloud.metadata

    def test_PointCloud_raw_dir_already_present(self):
        mymeta = gen_foldering_meta(
            os.path.join(TESTS_DIR, pointcloud_toy_dataset, "raw"),
            root=os.path.join(TESTS_DIR, pointcloud_toy_dataset),
        )
        a = PointCloud(metadata=mymeta, clear_processed=True)
        print(a)

    def test_PointCloud_sample_with_one_point_no_transform(self):
        mymeta = gen_foldering_meta(os.path.join(TESTS_DIR, empty_sample_dataset))
        a = PointCloud(metadata=mymeta, clear_processed=True)
        # print(a)
        for sample_x in a:
            # print(sample_x)
            assert isinstance(sample_x.y, int)
        b = PointCloud(metadata=mymeta)
        for sample_x in b:
            # print(sample_x)
            assert isinstance(sample_x.y, int)

    def test_PointCloud_sample_with_one_point_transform(self):
        # This is a pytorch-geometric/FixedPoints issue, see
        #   https://github.com/rusty1s/pytorch_geometric/issues/1090
        mymeta = gen_foldering_meta(os.path.join(TESTS_DIR, empty_sample_dataset))
        random_select_n_points = FixedPoints(num=100)
        a = PointCloud(
            metadata=mymeta, clear_processed=True, transform=random_select_n_points
        )
        for sample_x in a:
            print(sample_x)
            assert isinstance(sample_x.y, int)

    def test_PointCloud_different_raw_dirs_without_root(self):
        mymeta_toy = gen_foldering_meta(os.path.join(TESTS_DIR, pointcloud_toy_dataset))
        pointcloud_toy = PointCloud(metadata=mymeta_toy, clear_processed=True)
        mymeta_tcell = gen_foldering_meta(os.path.join(TESTS_DIR, tcell_toy_dataset))
        pointcloud_tcell = PointCloud(metadata=mymeta_tcell, clear_processed=True)
        assert (
            pointcloud_toy != pointcloud_tcell
        )  # the two pointclouds MUST not be the same.

        pointcloud_toy = PointCloud(metadata=mymeta_toy)
        # Even after loading pointcloud_toy a second time, the data MUST NOT be the same
        if pointcloud_toy[0].pos.shape == pointcloud_tcell[0].pos.shape:
            assert not torch.allclose(pointcloud_toy[0].pos, pointcloud_tcell[0].pos)

    def test_PointCloud_different_featurenames(self):
        mymeta = gen_foldering_meta(
            os.path.join(
                TESTS_DIR,
                pointcloud_toy_dataset,
                os.pardir,
                "Tcell_foldering_corrupted_colnames",
            ),
            root=os.path.join(TESTS_DIR, "new_datasets_dir"),
        )
        import copy

        saved_meta = copy.deepcopy(mymeta)
        assert saved_meta == mymeta
        with self.assertRaises(ValueError):
            my_pc = PointCloud(metadata=mymeta, clear_processed=True)
            repr(my_pc)
        assert saved_meta == mymeta

    def test_PointCloud_with_transform(self):
        mymeta = gen_foldering_meta(
            os.path.join(TESTS_DIR, pointcloud_toy_dataset, "raw")
        )
        random_select_n_points = FixedPoints(num=2)
        import numpy as np

        np.random.seed(123)
        a = PointCloud(mymeta, transform=random_select_n_points, clear_processed=True)
        target_tensor = torch.tensor(
            [
                [
                    15.8737,
                    20.8384,
                    15.6227,
                    8.6955,
                    8.7534,
                    8.2956,
                    9.0078,
                    7.8170,
                    8.1263,
                    8.9850,
                    7.2125,
                    8.6033,
                    11.7536,
                    4107.0000,
                ],
                [
                    15.5161,
                    20.8384,
                    14.9606,
                    8.4246,
                    8.7429,
                    7.6892,
                    9.5300,
                    -5.6333,
                    11.7794,
                    8.1770,
                    6.9872,
                    8.1666,
                    12.4236,
                    2882.0000,
                ],
            ]
        )
        assert torch.allclose(a[0].pos, target_tensor, atol=1e-3, rtol=1e-3)

    def test_PointCloud_FixedPoints_get_twice(self):
        mymeta = gen_foldering_meta(
            os.path.join(TESTS_DIR, pointcloud_toy_dataset, "raw")
        )
        random_select_n_points = FixedPoints(num=2)
        import numpy as np

        np.random.seed(123)
        a = PointCloud(mymeta, transform=random_select_n_points, clear_processed=True)

        # Even the same a[N] called twice gives different results!!!
        # This is because inside, transform is called every time with a get() call.
        assert not torch.allclose(a[0].pos, a[0].pos, rtol=1e-4)

        # in contrast, when setting the seed for the RNG inside, the following results must be the same.
        # God, this is tricky.
        np.random.seed(123)
        tmp1 = a[0]
        np.random.seed(123)
        tmp2 = a[0]
        assert torch.allclose(tmp1.pos, tmp2.pos, rtol=1e-4)

    def test_PointCloud_DataLoader(self):
        mymeta = gen_foldering_meta(
            os.path.join(TESTS_DIR, pointcloud_toy_dataset, "raw")
        )
        random_select_n_points = FixedPoints(num=2)
        import numpy as np

        np.random.seed(123)
        a = PointCloud(mymeta, transform=random_select_n_points, clear_processed=True)

        # # The torch DataLoader cannot work anymore (2020-02-26) because I switched my PointCloud dataset
        # # completely to be based on torch_geometric.data.data.Data
        # t_loader = t_DataLoader(a)
        # for batch in t_loader:
        #     print(batch)
        #     break

        tg_loader = tg_DataLoader(a)
        for batch in tg_loader:
            repr(batch)
            print(batch)

    def test_PointCloud_DataLoader_more_batches_than_samples(self):
        mymeta = gen_foldering_meta(
            os.path.join(TESTS_DIR, pointcloud_toy_dataset, "raw")
        )
        n_points_per_draw = 2
        random_select_n_points = FixedPoints(num=n_points_per_draw)
        import numpy as np

        np.random.seed(123)
        a = PointCloud(mymeta, transform=random_select_n_points, clear_processed=True)

        # Default torch_geometric dataloader just selects the maximum possible amount of samples.
        tg_loader = tg_DataLoader(
            a, batch_size=len(mymeta) + 1
        )  # Draw more samples than in mymeta present
        for batch in tg_loader:
            assert batch.pos.shape[0] / len(batch.y) == n_points_per_draw
            assert len(batch.y) == len(mymeta)

        # torch_geometric.loader bases on torch.utils.data.DataLoader
        # therefore we can supply a "sampler"
        # The sampler defines the complete "sample-set" which can be drawn from during learning.
        # The batch_size splits this "sample-set" and iterating over the DataLoader then has
        #   num_samples / batch_size elements
        tg_loader = tg_DataLoader(
            a,
            batch_size=100,
            sampler=torch.utils.data.RandomSampler(
                data_source=a, replacement=True, num_samples=100
            ),
        )  # Draw more samples than in mymeta present
        for batch in tg_loader:
            assert batch.pos.shape[0] / len(batch.y) == n_points_per_draw
            assert len(batch.y) == 100

        tg_loader = tg_DataLoader(
            a,
            batch_size=20,
            sampler=torch.utils.data.RandomSampler(
                data_source=a, replacement=True, num_samples=100
            ),
        )  # Draw more samples than in mymeta present
        # 100 samples sampled randomly in total
        # 20 per batchsize
        # --> I have 5 batches
        assert len(tg_loader) == 5
        for batch in tg_loader:
            assert batch.pos.shape[0] / len(batch.y) == n_points_per_draw
            assert len(batch.y) == 20

        # What if the batch_size does not exactly part num_samples?
        tg_loader = tg_DataLoader(
            a,
            batch_size=20,
            sampler=torch.utils.data.RandomSampler(
                data_source=a, replacement=True, num_samples=90
            ),
        )  # Draw more samples than in mymeta present
        # 90 samples sampled randomly in total
        # 20 per batchsize
        # --> I have 5 batches and the last batch has just 10 samples
        assert len(tg_loader) == 5
        lenghts = []
        for batch in tg_loader:
            lenghts.append(len(batch.y))
            # assert batch.pos.shape[0] / len(batch.y) == n_points_per_draw
            # assert len(batch.y) == 20
        assert lenghts == [20, 20, 20, 20, 10]

    def tearDown(self) -> None:
        try:
            shutil.rmtree(os.path.join(TESTS_DIR, pointcloud_toy_dataset, "processed"))
            pass
        except FileNotFoundError:
            pass
        try:
            shutil.rmtree(
                os.path.join(TESTS_DIR, pointcloud_toy_dataset, "raw", "processed")
            )
            shutil.rmtree(os.path.join(TESTS_DIR, pointcloud_toy_dataset, "raw", "raw"))
            pass
        except FileNotFoundError:
            pass
        try:
            shutil.rmtree(os.path.join(TESTS_DIR, "new_datasets_dir"))
        except FileNotFoundError:
            pass



from typing import List, Tuple

import pandas as pd
import torch

try:
    import datatable

    def csv_reader(*args, **kwargs):
        read_data = datatable.fread(*args, **kwargs)
        return torch.tensor(read_data.to_numpy()), list(read_data.names)

except ImportError:
    # fallback to pandas
    def csv_reader(*args, **kwargs):
        read_data = pd.read_csv(*args, **kwargs)
        return torch.tensor(read_data.values), [x.strip() for x in read_data.columns]


def read_csv(
    path: str, sep: str = ",", *args, **kwargs
) -> Tuple[torch.Tensor, List[str]]:
    """
    read_csv reads the csv
    - first line are assumed to be feature names.
    - All further lines must be numeric

    Returns:
         Tuple of point cloud data and the column names

    """
    values_torch, colnames_list = csv_reader(path, *args, sep=sep, **kwargs)
    return values_torch, colnames_list



import os
from unittest import TestCase

import torch

from pointcyto.data.InMemoryPointCloud import InMemoryPointCloud
from pointcyto.io.meta_read_foldering import gen_foldering_meta
from pointcyto.testutils.helpers import find_dirname_above_currentfile
from pointcyto.transforms.transforms import (
    CalculateOutlierness,
    QminmaxPointCloud,
    global_mean,
    global_sd,
    qminmax,
)

THIS_DIR = os.path.dirname(os.path.abspath(__file__))
# Now build the base-test directory (necessary if you want to use tests/testdata)
TESTS_DIR = find_dirname_above_currentfile()

pointcloud_toy_dataset = os.path.join(
    "testdata", "flowcytometry", "PointClouds_toy_dataset"
)

pointcloud_toy_dataset_mean = torch.tensor(
    [
        15.1515,
        20.7125,
        15.1940,
        8.9993,
        8.7140,
        7.5958,
        7.6895,
        6.9978,
        8.0298,
        5.7519,
        7.8421,
        8.2171,
        11.3506,
        4120.7495,
    ]
)

pointcloud_toy_dataset_sd = torch.tensor(
    [
        0.71723,
        0.21693,
        1.0523,
        1.5938,
        2.4265,
        4.7020,
        5.0719,
        5.7231,
        3.9511,
        6.3471,
        2.7981,
        3.7452,
        3.0114,
        2554.4,
    ]
)


class TestTransforms(TestCase):
    def test_global_mean(self):
        mymeta = gen_foldering_meta(
            os.path.join(TESTS_DIR, pointcloud_toy_dataset, "raw")
        )
        complete_data = InMemoryPointCloud(mymeta, clear_processed=True)
        assert torch.allclose(
            global_mean(complete_data),
            pointcloud_toy_dataset_mean,
            rtol=1e-3,
        )

    def test_global_sd(self):
        mymeta = gen_foldering_meta(
            os.path.join(TESTS_DIR, pointcloud_toy_dataset, "raw")
        )
        complete_data = InMemoryPointCloud(mymeta, clear_processed=True)
        assert torch.allclose(
            global_sd(complete_data),
            pointcloud_toy_dataset_sd,
            rtol=1e-3,
        )

    def test_global_sd_precalculate_mean(self):
        mymeta = gen_foldering_meta(
            os.path.join(TESTS_DIR, pointcloud_toy_dataset, "raw")
        )
        complete_data = InMemoryPointCloud(mymeta, clear_processed=True)
        means = global_mean(complete_data)
        assert torch.allclose(
            global_sd(complete_data, global_means=means),
            pointcloud_toy_dataset_sd,
            rtol=1e-3,
        )

    def test_NormalizePointCloud(self):
        from pointcyto.transforms.transforms import NormalizePointCloud

        repr(NormalizePointCloud(mean=torch.zeros((14)), std=torch.ones((14))))
        mymeta = gen_foldering_meta(
            os.path.join(TESTS_DIR, pointcloud_toy_dataset, "raw")
        )

        no_transformation = InMemoryPointCloud(
            mymeta,
            clear_processed=True,
        )
        # You cannot use NormalizePointCloud without defining mean and std.
        with self.assertRaises(TypeError):
            complete_data = InMemoryPointCloud(
                mymeta, clear_processed=True, pre_transform=NormalizePointCloud()
            )
            print(complete_data)

        transformation_but_useless = InMemoryPointCloud(
            mymeta,
            clear_processed=True,
            pre_transform=NormalizePointCloud(
                mean=torch.zeros((14)), std=torch.ones((14))
            ),
        )

        transformation_halved = InMemoryPointCloud(
            mymeta,
            clear_processed=True,
            pre_transform=NormalizePointCloud(
                mean=torch.zeros((14)), std=torch.ones((14)) * 2
            ),
        )
        assert torch.allclose(no_transformation.pos, transformation_but_useless.pos)
        assert torch.allclose(no_transformation.pos, transformation_halved.pos * 2)

    def test_WeightedFixedPoints_repr(self):
        from pointcyto.transforms.transforms import WeightedFixedPoints

        a = WeightedFixedPoints(10, 5)
        repr(a)

    def test_CalculateOutlierness_repr(self):
        from pointcyto.transforms.transforms import CalculateOutlierness

        a = CalculateOutlierness(10)
        repr(a)

    def test_CalculateOutlierness(self):
        mymeta = gen_foldering_meta(
            os.path.join(TESTS_DIR, pointcloud_toy_dataset, "raw")
        )
        complete_data = InMemoryPointCloud(
            mymeta,
            pre_transform=CalculateOutlierness(n_reference=500, use_cuda=False),
            clear_processed=True,
        )
        assert "outlierness" in complete_data._data.keys()
        assert complete_data.outlierness.shape[0] == complete_data.pos.shape[0]

    def test_CalculateOutlierness_with_without_cuda_checkType(self):
        mymeta = gen_foldering_meta(
            os.path.join(TESTS_DIR, pointcloud_toy_dataset, "raw")
        )
        complete_data = InMemoryPointCloud(
            mymeta,
            pre_transform=CalculateOutlierness(n_reference=500, use_cuda=False),
            clear_processed=True,
        )
        assert "outlierness" in complete_data._data.keys()
        assert complete_data.outlierness.shape[0] == complete_data.pos.shape[0]

        if not torch.cuda.is_available():
            pass
        else:
            complete_data_cuda = InMemoryPointCloud(
                mymeta,
                pre_transform=CalculateOutlierness(n_reference=500, use_cuda=True),
                clear_processed=True,
            )
            assert "outlierness" in complete_data._data.keys()
            assert complete_data.outlierness.shape[0] == complete_data_cuda.pos.shape[0]

            assert not complete_data.outlierness.is_cuda
            assert complete_data_cuda.outlierness.is_cuda
            # The following assertion must be false because The 500 points in both cases drawn ar different.
            assert not torch.allclose(
                complete_data.outlierness,
                complete_data_cuda.outlierness.cpu(),
            )

    def test_CalculateOutlierness_with_without_cuda_check_euqal_cpu(self):
        import numpy as np

        from pointcyto.transforms.transform_param_onlist import NormalizePointCloudParam

        mymeta = gen_foldering_meta(
            os.path.join(TESTS_DIR, pointcloud_toy_dataset, "raw")
        )
        np.random.seed(164)
        complete_data = InMemoryPointCloud(
            mymeta,
            pre_pre_transform_param_onlist=NormalizePointCloudParam(),
            pre_transform=CalculateOutlierness(n_reference=1, use_cuda=False),
            clear_processed=True,
        )
        if not torch.cuda.is_available():
            pass
        else:
            np.random.seed(164)
            complete_data_cuda = InMemoryPointCloud(
                mymeta,
                pre_pre_transform_param_onlist=NormalizePointCloudParam(),
                pre_transform=CalculateOutlierness(n_reference=1, use_cuda=True),
                clear_processed=True,
            )

            assert torch.allclose(
                complete_data.outlierness,
                complete_data_cuda.outlierness.cpu(),
                atol=1e-1,
            )

    def test_CalculateOutlierness_in_batches(self):
        import numpy as np

        from pointcyto.transforms.transform_param_onlist import NormalizePointCloudParam

        mymeta = gen_foldering_meta(
            os.path.join(TESTS_DIR, pointcloud_toy_dataset, "raw")
        )
        np.random.seed(164)
        complete_data_10 = InMemoryPointCloud(
            mymeta,
            pre_pre_transform_param_onlist=NormalizePointCloudParam(),
            pre_transform=CalculateOutlierness(
                n_reference=10, use_cuda=False, calculate_in_batches=10
            ),
            clear_processed=True,
        )
        np.random.seed(164)
        complete_data_1 = InMemoryPointCloud(
            mymeta,
            pre_pre_transform_param_onlist=NormalizePointCloudParam(),
            pre_transform=CalculateOutlierness(
                n_reference=10, use_cuda=False, calculate_in_batches=1
            ),
            clear_processed=True,
        )
        np.random.seed(164)
        complete_data_None = InMemoryPointCloud(
            mymeta,
            pre_pre_transform_param_onlist=NormalizePointCloudParam(),
            pre_transform=CalculateOutlierness(n_reference=10, use_cuda=False),
            clear_processed=True,
        )
        assert torch.allclose(
            complete_data_1[0]["outlierness"], complete_data_None[0]["outlierness"]
        )
        assert torch.allclose(
            complete_data_1[0]["outlierness"],
            complete_data_10[0]["outlierness"],
            atol=1e3,
        )

    def test_CalculateOutlierness_drawMorePointsThanPresent(self):
        mymeta = gen_foldering_meta(
            os.path.join(TESTS_DIR, pointcloud_toy_dataset, "raw")
        )
        import torch_geometric as tg

        from pointcyto.transforms.transforms import FixedPointsOutlierness

        with self.assertRaises(ValueError):
            # Too little number of points, cannot draw 500k from each sample.
            complete_data = InMemoryPointCloud(
                mymeta,
                pre_transform=tg.transforms.compose.Compose(
                    [
                        CalculateOutlierness(n_reference=100, use_cuda=False),
                        FixedPointsOutlierness(num=500000, replace=False),
                    ]
                ),
                clear_processed=True,
            )

        complete_data = InMemoryPointCloud(
            mymeta,
            pre_transform=tg.transforms.compose.Compose(
                [
                    CalculateOutlierness(n_reference=100, use_cuda=False),
                    FixedPointsOutlierness(
                        num=500000, replace=False, allow_duplicates=True
                    ),
                ]
            ),
            clear_processed=True,
        )
        assert "outlierness" in complete_data._data.keys()
        assert complete_data.outlierness.shape[0] == complete_data.pos.shape[0]

    def test_qminmax_function(self):
        # 10 cells, 4 parameters
        new_tensor = torch.rand(size=(100, 4)) * 10
        print(new_tensor)
        print(new_tensor.shape)
        a = qminmax(tensor=new_tensor, quantile_range=[0.1, 0.9])
        print(a)

    def test_Qminmax_normalize(self):
        import numpy as np

        mymeta = gen_foldering_meta(
            os.path.join(TESTS_DIR, pointcloud_toy_dataset, "raw")
        )
        np.random.seed(164)
        complete_data_10 = InMemoryPointCloud(
            mymeta,
            pre_transform=QminmaxPointCloud(quantile_range=[0.01, 0.99]),
            clear_processed=True,
        )
        print(complete_data_10)



from typing import List, Tuple

import flowkit
import torch


def read_fcs(path: str) -> Tuple[torch.Tensor, List[str]]:
    """
    read_fcs extracts the original events *without* applying compensation, transformation or anything else.
    If you want to have compensated values, you must do this beforehand,
    - store it into '_orig_events'
    - export it into the new .fcs

    Example:

        .. code-block::

            import flowkit

            orig = flowkit.Sample('removeme_original.fcs')
            a = orig.get_orig_events()
            a[:, :] = 0  # compensate here whatever you want
            orig._orig_events = a
            orig.export("removeme_zeroed_by_python.fcs")



    Args:
        path:
            Path to csv file. Must contain a head line
    Returns:
        Tuple of point cloud data and the column names
    :param path:

    Examples:

    .. code-block::

        print(read_fcs("data/Tcell/P0H.fcs"))

    """
    fk_sample = flowkit.Sample(path, cache_original_events=True)

    labels_pns = fk_sample.pns_labels
    labels_pnn = fk_sample.pnn_labels

    combined_labels = [
        value + "_" + labels_pnn[index] for index, value in enumerate(labels_pns)
    ]
    combined_labels_no_spaces = [
        namex.strip().replace(" ", "_") for namex in combined_labels
    ]
    return torch.from_numpy(fk_sample._orig_events), combined_labels_no_spaces



from typing import List, Tuple

import torch

try:
    # https://arrow.apache.org/docs/python/generated/pyarrow.feather.read_feather.html
    import pyarrow.feather as feather

    def feather_reader(*args, **kwargs):
        read_data = feather.read_feather(*args, **kwargs)
        return torch.tensor(read_data.to_numpy()), list(read_data.columns)

    # Alternatively could use pandas, but pandas just calls pyarrow.
    # # https://pandas.pydata.org/docs/reference/api/pandas.read_feather.html
    # import pandas as pd
    # # Pandas fallback.
    # # 2023-03-13: I am not sure if pandas or pyarrow is better, but if pyarrow is installed it is
    # # used.
    # def feather_reader(*args, **kwargs):
    #     read_data = pd.read_feather(*args, **kwargs)
    #     return torch.tensor(read_data.to_numpy()), list(read_data.columns)

except ImportError:
    # No fallback
    def feather_reader(*args, **kwargs):
        raise ImportError("Did not find pyarrow, cannot read feather files")


def read_feather(path: str, *args, **kwargs) -> Tuple[torch.Tensor, List[str]]:
    """
    read_feather reads the feather file
    - first line are assumed to be feature names.
    - All further lines must be numeric

    Returns:
         Tuple of point cloud data and the column names

    """
    values_torch, colnames_list = feather_reader(path, *args, **kwargs)
    return values_torch, colnames_list



from metadata_basetests import BaseTests

from pointcyto.data.metadata_classification import MetaDataClassification


class TestMetaDataClassification(BaseTests):
    metadata = MetaDataClassification



from typing import List, Tuple

import torch

try:
    import datatable

    def jay_reader(*args, **kwargs):
        read_data = datatable.fread(*args, **kwargs)
        return torch.tensor(read_data.to_numpy()), list(read_data.names)

except ImportError:
    # No fallback
    def jay_reader(*args, **kwargs):
        raise ImportError("Did not find datatable, cannot read jay files")


def read_jay(path: str, *args, **kwargs) -> Tuple[torch.Tensor, List[str]]:
    """
    read_jay reads the jay file (saved by datatable usually)

    Returns:
         Tuple of point cloud data and the column names

    """
    values_torch, colnames_list = jay_reader(path, *args, **kwargs)
    return values_torch, colnames_list



import pickle


def pickle_open_dump(obj, path: str, verbose: bool = True) -> None:
    """

    Args:
        obj:
            The object to pickle
        path:
            Where the pickle should be saved
        verbose:
            If verbose the output path is printed.

    Returns:

    """
    with open(path, "wb") as f:
        pickle.dump(obj, f)
    if verbose:
        print("Saved: " + path)



import os
import shutil
from unittest import TestCase

import torch
from torch_geometric.transforms.compose import Compose

from pointcyto.data.InMemoryPointCloud import InMemoryPointCloud
from pointcyto.io.meta_read_foldering import gen_foldering_meta
from pointcyto.testutils.helpers import find_dirname_above_currentfile
from pointcyto.transforms.transform_param_onlist import (
    FixedPointsOnList,
    NormalizePointCloudParam,
)

TESTS_DIR = find_dirname_above_currentfile()


pointcloud_toy_dataset = os.path.join(
    "testdata", "flowcytometry", "PointClouds_toy_dataset"
)


class TestPrePreTransformsOnList(TestCase):
    def test_NormalizePointCloudParam_as_list_and_as_single_transform(self):
        mymeta = gen_foldering_meta(
            os.path.join(TESTS_DIR, pointcloud_toy_dataset, "raw")
        )
        complete_data = InMemoryPointCloud(
            mymeta,
            pre_pre_transform_param_onlist=[NormalizePointCloudParam()],
            clear_processed=True,
        )
        complete_data = InMemoryPointCloud(
            mymeta,
            pre_pre_transform_param_onlist=NormalizePointCloudParam(),
            clear_processed=True,
        )
        print(complete_data)

    def test_NormalizePointCloudParam_multiple(self):
        mymeta = gen_foldering_meta(
            os.path.join(TESTS_DIR, pointcloud_toy_dataset, "raw")
        )
        complete_data = InMemoryPointCloud(
            mymeta,
            pre_pre_transform_param_onlist=[
                NormalizePointCloudParam(),
                NormalizePointCloudParam(),
            ],
            clear_processed=True,
        )
        repr(complete_data)

    def test_NormalizePointCloudParam_faulty_in_pre_transform(self):
        mymeta = gen_foldering_meta(
            os.path.join(TESTS_DIR, pointcloud_toy_dataset, "raw")
        )
        with self.assertRaises(TypeError):
            complete_data = InMemoryPointCloud(
                mymeta,
                pre_transform=[NormalizePointCloudParam(), NormalizePointCloudParam()],
                clear_processed=True,
            )
        with self.assertRaises(AttributeError):
            complete_data = InMemoryPointCloud(
                mymeta, pre_transform=NormalizePointCloudParam(), clear_processed=True
            )
            print(complete_data)

    def test_NormalizePointCloudParam_pre_pre_transform_param_list_save_load(self):
        from pointcyto.transforms.transform_param_onlist import NormalizePointCloudParam

        mymeta = gen_foldering_meta(
            os.path.join(TESTS_DIR, pointcloud_toy_dataset, "raw")
        )
        prepre_trans_list = [NormalizePointCloudParam(), NormalizePointCloudParam()]
        mydataset = InMemoryPointCloud(
            mymeta,
            pre_pre_transform_param_onlist=prepre_trans_list,
            clear_processed=True,
        )

        import pickle

        with open("pretransform_parameter.pickle", "wb") as f:
            pickle.dump(mydataset.pretransform_parameter, f)
        with open("pretransform_parameter.pickle", "rb") as f:
            loaded_pretransform_parameter = pickle.load(f)

        transform_list = []
        for single_pretransform in loaded_pretransform_parameter:
            recreated_transform = single_pretransform["related_transform"](
                **single_pretransform["param"]
            )
            with self.assertRaises(TypeError):
                repr(single_pretransform["related_transform"]())
            transform_list += [recreated_transform]
        composed_transform_list = Compose(transform_list)

        with self.assertWarns(UserWarning):
            # Here a UserWarning is issues because we use the same mymeta but different transforms as before
            # This is absolutely expected behaviour
            mydataset_reloaded = InMemoryPointCloud(
                mymeta, pre_transform=composed_transform_list
            )

        import warnings

        with warnings.catch_warnings(record=True) as w:
            # In contrast to before, here we clear the Pointcloud related to mymeta, so no warning must occur
            mydataset_reloaded = InMemoryPointCloud(
                mymeta, pre_transform=composed_transform_list, clear_processed=True
            )
            if any(issubclass(w_.category, UserWarning) for w_ in w):
                raise AssertionError("UserWarning was issued")

        assert torch.allclose(mydataset_reloaded[0].pos, mydataset[0].pos)

    def test_NormalizePointCloudParam_FixedPoints(self):
        mymeta = gen_foldering_meta(
            os.path.join(TESTS_DIR, pointcloud_toy_dataset, "raw")
        )
        complete_data = InMemoryPointCloud(
            mymeta,
            pre_pre_transform_param_onlist=[
                FixedPointsOnList(num=100),
                NormalizePointCloudParam(),
            ],
            clear_processed=True,
        )
        repr(complete_data)
        assert complete_data.pos.shape[0] == 600

    def test_ScaleLowHigh01(self):
        from pointcyto.transforms.transform_param_onlist import ScaleLowHigh01OnList

        mymeta = gen_foldering_meta(
            os.path.join(TESTS_DIR, pointcloud_toy_dataset, "raw")
        )
        original_data = InMemoryPointCloud(
            mymeta,
            clear_processed=True,
        )
        last_pos_value = original_data.pos[-1, -1]

        a = InMemoryPointCloud(
            mymeta,
            pre_pre_transform_param_onlist=[
                ScaleLowHigh01OnList(),
            ],
            clear_processed=True,
        )
        repr(a)

        # one low-high shift for all samples
        a = InMemoryPointCloud(
            mymeta,
            pre_pre_transform_param_onlist=[
                # ScaleLowHigh01OnList([[[0, tensor(6406.0)]]]),
                ScaleLowHigh01OnList([[[0, last_pos_value]]]),
            ],
            clear_processed=True,
        )

        assert a.pos[-1, -1] == 1
        # A low-high shift per feature for all samples
        with self.assertRaises(ValueError):
            # Wrong brackets multiplied
            a = InMemoryPointCloud(
                mymeta,
                pre_pre_transform_param_onlist=[
                    ScaleLowHigh01OnList([[[0, last_pos_value] * 14]]),
                ],
                clear_processed=True,
            )
        a = InMemoryPointCloud(
            mymeta,
            pre_pre_transform_param_onlist=[
                ScaleLowHigh01OnList([[[0, last_pos_value]] * 14]),
            ],
            clear_processed=True,
        )
        assert a.pos[-1, -1] == 1
        with self.assertRaises(ValueError):
            a = InMemoryPointCloud(
                mymeta,
                pre_pre_transform_param_onlist=[
                    ScaleLowHigh01OnList([[[0, last_pos_value]] * 13]),
                ],
                clear_processed=True,
            )
        a = InMemoryPointCloud(
            mymeta,
            pre_pre_transform_param_onlist=[
                ScaleLowHigh01OnList([[[1]] * 14]),
            ],
            clear_processed=True,
        )
        assert torch.allclose(original_data.pos - a.pos, torch.tensor(1.0))
        a = InMemoryPointCloud(
            mymeta,
            pre_pre_transform_param_onlist=[
                ScaleLowHigh01OnList([[[1]]]),
            ],
            clear_processed=True,
        )
        assert torch.allclose(original_data.pos - a.pos, torch.tensor(1.0))
        a = InMemoryPointCloud(
            mymeta,
            pre_pre_transform_param_onlist=[
                ScaleLowHigh01OnList([[[i] for i in range(14)]]),
            ],
            clear_processed=True,
        )
        assert torch.allclose(
            (original_data.pos - a.pos)[1, :],
            torch.tensor(
                [
                    0.0,
                    1.0,
                    2.0,
                    3.0,
                    4.0,
                    5.0,
                    6.0,
                    7.0,
                    8.0,
                    9.0,
                    10.0,
                    11.0,
                    12.0,
                    13.0,
                ]
            ),
        )
        assert torch.allclose(
            (original_data.pos - a.pos)[-1, :],
            torch.tensor(
                [
                    0.0,
                    1.0,
                    2.0,
                    3.0,
                    4.0,
                    5.0,
                    6.0,
                    7.0,
                    8.0,
                    9.0,
                    10.0,
                    11.0,
                    12.0,
                    13.0,
                ]
            ),
        )

    def test_ReTransformShiftAsinhOnList(self):
        try:
            shutil.rmtree("removeme")
        except FileNotFoundError:
            pass
        from pointcyto.transforms.transform_param_onlist import (
            ReTransformShiftAsinhOnList,
        )

        mymeta = gen_foldering_meta(
            os.path.join(TESTS_DIR, pointcloud_toy_dataset, "raw")
        )
        original_data = InMemoryPointCloud(
            mymeta,
            clear_processed=True,
        )
        # Apply on already asinh-transformed values
        from pointcyto.transforms.transforms_flowcore import ArcsinhTransform

        asinh_transformed = InMemoryPointCloud(
            mymeta,
            pre_transform=ArcsinhTransform(
                a=0,
                b=1 / 5.12,
                c=0,
            ),
            clear_processed=True,
        )
        asinh_transformed.to_fcs("removeme")
        asinh_trans_meta = gen_foldering_meta("removeme")
        with self.assertWarns(UserWarning):
            # UserWarning: Do not use this transformation on transformed AND
            # shifted values, only on transformed and NOT shifted.
            asinh_RE_transformed = InMemoryPointCloud(
                asinh_trans_meta,
                pre_pre_transform_param_onlist=[
                    ReTransformShiftAsinhOnList(
                        # dataset_attribute="pos",
                        samplewise_feature_shifts_asinhspace=[[[0]]],
                        do_shift_in_rawspace=False,
                        original_asinh_cofactors=[5.12],
                        original_asinh_cofactors_applied=True,
                        new_asinh_cofactors=[5.12],
                        inplace=True,
                    ),
                ],
                clear_processed=True,
            )
        assert torch.allclose(asinh_transformed.pos, asinh_RE_transformed.pos)

        with self.assertRaises(ValueError):
            # "100" is a totally unreasonable value after asinh transformation
            a = InMemoryPointCloud(
                mymeta,
                pre_pre_transform_param_onlist=[
                    ReTransformShiftAsinhOnList(
                        # dataset_attribute="pos",
                        samplewise_feature_shifts_asinhspace=[[[1, 100]]],
                        do_shift_in_rawspace=False,
                        original_asinh_cofactors=[1],
                        original_asinh_cofactors_applied=False,
                        new_asinh_cofactors=[1],
                        inplace=True,
                    ),
                ],
                clear_processed=True,
            )

        with self.assertRaises(ValueError):
            # Different length of cofactors
            a = InMemoryPointCloud(
                mymeta,
                pre_pre_transform_param_onlist=[
                    ReTransformShiftAsinhOnList(
                        # dataset_attribute="pos",
                        samplewise_feature_shifts_asinhspace=[[[0]]],
                        do_shift_in_rawspace=False,
                        original_asinh_cofactors=[1],
                        original_asinh_cofactors_applied=False,
                        new_asinh_cofactors=[1] * 14,
                        inplace=True,
                    ),
                ],
                clear_processed=True,
            )
        a = InMemoryPointCloud(
            mymeta,
            pre_pre_transform_param_onlist=[
                ReTransformShiftAsinhOnList(
                    # dataset_attribute="pos",
                    samplewise_feature_shifts_asinhspace=[[[0]]],
                    do_shift_in_rawspace=False,
                    original_asinh_cofactors=[1],
                    original_asinh_cofactors_applied=False,
                    new_asinh_cofactors=[1],
                    inplace=True,
                ),
            ],
            clear_processed=True,
        )
        assert torch.allclose(a.pos, torch.asinh(original_data.pos))

        with self.assertRaises(ValueError):
            a = InMemoryPointCloud(
                mymeta,
                pre_pre_transform_param_onlist=[
                    ReTransformShiftAsinhOnList(
                        # dataset_attribute="pos",
                        samplewise_feature_shifts_asinhspace=[[[0]]],
                        do_shift_in_rawspace=False,
                        original_asinh_cofactors=[1] * 14,
                        original_asinh_cofactors_applied=False,
                        new_asinh_cofactors=[1] * 14,
                        inplace=True,
                    ),
                ],
                clear_processed=True,
            )
        with self.assertRaises(ValueError):
            # Wrong bracket multiplied
            a = InMemoryPointCloud(
                mymeta,
                pre_pre_transform_param_onlist=[
                    ReTransformShiftAsinhOnList(
                        # dataset_attribute="pos",
                        samplewise_feature_shifts_asinhspace=[[[0] * 14]],
                        do_shift_in_rawspace=False,
                        original_asinh_cofactors=[1] * 14,
                        original_asinh_cofactors_applied=False,
                        new_asinh_cofactors=[1] * 14,
                        inplace=True,
                    ),
                ],
                clear_processed=True,
            )
        a = InMemoryPointCloud(
            mymeta,
            pre_pre_transform_param_onlist=[
                ReTransformShiftAsinhOnList(
                    # dataset_attribute="pos",
                    samplewise_feature_shifts_asinhspace=[[[0]] * 14],
                    do_shift_in_rawspace=False,
                    original_asinh_cofactors=[1] * 14,
                    original_asinh_cofactors_applied=False,
                    new_asinh_cofactors=[1] * 14,
                    inplace=True,
                ),
            ],
            clear_processed=True,
        )
        assert torch.allclose(a.pos, torch.asinh(original_data.pos))

        # Single and dual values per sample combined
        # print([[[0]] * 13 + [[1, 3]]])
        a = InMemoryPointCloud(
            mymeta,
            pre_pre_transform_param_onlist=[
                ReTransformShiftAsinhOnList(
                    # dataset_attribute="pos",
                    samplewise_feature_shifts_asinhspace=[[[0]] * 13 + [[1, 3]]],
                    do_shift_in_rawspace=False,
                    original_asinh_cofactors=[1] * 14,
                    original_asinh_cofactors_applied=False,
                    new_asinh_cofactors=[1] * 14,
                    inplace=True,
                ),
            ],
            clear_processed=True,
        )
        a = InMemoryPointCloud(
            mymeta,
            pre_pre_transform_param_onlist=[
                ReTransformShiftAsinhOnList(
                    # dataset_attribute="pos",
                    samplewise_feature_shifts_asinhspace=[[[1]]],
                    do_shift_in_rawspace=False,
                    original_asinh_cofactors=[1],
                    original_asinh_cofactors_applied=False,
                    new_asinh_cofactors=[1],
                    inplace=True,
                ),
            ],
            clear_processed=True,
        )
        assert torch.allclose(a.pos, torch.asinh(original_data.pos) - 1, atol=1e-4)
        a = InMemoryPointCloud(
            mymeta,
            pre_pre_transform_param_onlist=[
                ReTransformShiftAsinhOnList(
                    # dataset_attribute="pos",
                    samplewise_feature_shifts_asinhspace=[[[1, 4]]],
                    do_shift_in_rawspace=False,
                    original_asinh_cofactors=[1],
                    original_asinh_cofactors_applied=False,
                    new_asinh_cofactors=[1],
                    inplace=True,
                ),
            ],
            clear_processed=True,
        )
        assert torch.allclose(
            a.pos, (torch.asinh(original_data.pos) - 1) / (4 - 1), atol=1e-4
        )
        a = InMemoryPointCloud(
            mymeta,
            pre_pre_transform_param_onlist=[
                ReTransformShiftAsinhOnList(
                    # dataset_attribute="pos",
                    samplewise_feature_shifts_asinhspace=[[[0]]],
                    do_shift_in_rawspace=False,
                    original_asinh_cofactors=[1],
                    original_asinh_cofactors_applied=False,
                    new_asinh_cofactors=[2],
                    inplace=True,
                ),
            ],
            clear_processed=True,
        )
        assert torch.allclose(a.pos, torch.asinh(original_data.pos / 2.0), atol=1e-4)

        a = InMemoryPointCloud(
            mymeta,
            pre_pre_transform_param_onlist=[
                ReTransformShiftAsinhOnList(
                    # dataset_attribute="pos",
                    samplewise_feature_shifts_asinhspace=[[[1, 4]] * 14],
                    do_shift_in_rawspace=False,
                    original_asinh_cofactors=[1] * 14,
                    original_asinh_cofactors_applied=False,
                    new_asinh_cofactors=[2] * 14,
                    inplace=True,
                ),
            ],
            clear_processed=True,
        )
        assert torch.allclose(
            a.pos,
            (
                (
                    torch.asinh(original_data.pos / 2.0)
                    - (torch.asinh(torch.sinh(torch.tensor(1)) / 2))
                )
                / (
                    torch.asinh(torch.sinh(torch.tensor(4)) / 2)
                    - torch.asinh(torch.sinh(torch.tensor(1)) / 2)
                )
            ),
            atol=1e-4,
        )
        a = InMemoryPointCloud(
            mymeta,
            pre_pre_transform_param_onlist=[
                ReTransformShiftAsinhOnList(
                    # dataset_attribute="pos",
                    samplewise_feature_shifts_asinhspace=[[[1, 4]] * 14],
                    do_shift_in_rawspace=False,
                    original_asinh_cofactors=[3] * 14,
                    original_asinh_cofactors_applied=False,
                    new_asinh_cofactors=[2] * 14,
                    inplace=True,
                ),
            ],
            clear_processed=True,
        )
        assert torch.allclose(
            a.pos,
            (
                (
                    torch.asinh(original_data.pos / 2.0)
                    - (torch.asinh(torch.sinh(torch.tensor(1)) / 2 * 3))
                )
                / (
                    torch.asinh(torch.sinh(torch.tensor(4)) / 2 * 3)
                    - torch.asinh(torch.sinh(torch.tensor(1)) / 2 * 3)
                )
            ),
            atol=1e-4,
        )

        a = InMemoryPointCloud(
            mymeta,
            pre_pre_transform_param_onlist=[
                ReTransformShiftAsinhOnList(
                    # dataset_attribute="pos",
                    samplewise_feature_shifts_asinhspace=[[[0]] * 14],
                    do_shift_in_rawspace=False,
                    original_asinh_cofactors=[x + 1 for x in range(14)],
                    original_asinh_cofactors_applied=False,
                    new_asinh_cofactors=[x + 1 for x in range(14)],
                    inplace=True,
                ),
            ],
            clear_processed=True,
        )
        from pointcyto.transforms.transforms_functional import scale_featurewise

        assert torch.allclose(
            a.pos,
            torch.asinh(
                scale_featurewise(
                    original_data.pos,
                    feature_scaling_vector=[x + 1 for x in range(14)],
                )
            ),
        )

        a = InMemoryPointCloud(
            mymeta,
            pre_pre_transform_param_onlist=[
                ReTransformShiftAsinhOnList(
                    # dataset_attribute="pos",
                    samplewise_feature_shifts_asinhspace=[[[0]] * 14],
                    do_shift_in_rawspace=False,
                    original_asinh_cofactors=[x + 1 for x in range(14)],
                    original_asinh_cofactors_applied=False,
                    new_asinh_cofactors=[x + 2 for x in range(14)],
                    inplace=True,
                ),
            ],
            clear_processed=True,
        )
        assert torch.allclose(
            a.pos,
            torch.asinh(
                scale_featurewise(
                    original_data.pos,
                    feature_scaling_vector=[x + 2 for x in range(14)],
                )
            ),
        )

        old_cof = [x + 1 for x in range(14)]
        new_cof = [x * 2 + 5 for x in range(14)]
        a = InMemoryPointCloud(
            mymeta,
            pre_pre_transform_param_onlist=[
                ReTransformShiftAsinhOnList(
                    # dataset_attribute="pos",
                    samplewise_feature_shifts_asinhspace=[[[1, 4]] * 14],
                    do_shift_in_rawspace=False,
                    original_asinh_cofactors=old_cof,
                    original_asinh_cofactors_applied=False,
                    new_asinh_cofactors=new_cof,
                    inplace=True,
                ),
            ],
            clear_processed=True,
        )
        for col_i in range(14):
            # assert torch.allclose(
            print(
                a.pos[:, col_i],
                (
                    (
                        torch.asinh(original_data.pos[:, col_i] / new_cof[col_i])
                        - (
                            torch.asinh(
                                torch.sinh(torch.tensor(1))
                                / new_cof[col_i]
                                * old_cof[col_i]
                            )
                        )
                    )
                    / (
                        torch.asinh(
                            torch.sinh(torch.tensor(4))
                            / new_cof[col_i]
                            * old_cof[col_i]
                        )
                        - torch.asinh(
                            torch.sinh(torch.tensor(1))
                            / new_cof[col_i]
                            * old_cof[col_i]
                        )
                    )
                ),
                # atol=1e-4
            )

    def tearDown(self) -> None:
        try:
            os.remove("pretransform_parameter.pickle")
            pass
        except FileNotFoundError:
            pass



import os
from unittest import TestCase

from pointcyto.data.InMemoryPointCloud import InMemoryPointCloud
from pointcyto.io.meta_read_foldering import gen_foldering_meta
from pointcyto.testutils.helpers import find_dirname_above_currentfile
from pointcyto.transforms.transforms_flowkit import FKtransform

THIS_DIR = os.path.dirname(os.path.abspath(__file__))
# Now build the base-test directory (necessary if you want to use tests/testdata)
TESTS_DIR = find_dirname_above_currentfile()

pointcloud_toy_dataset = os.path.join(
    "testdata", "flowcytometry", "PointClouds_toy_dataset"
)


class TestTransforms(TestCase):
    def test_AsinhTransform_repr(self):
        a = FKtransform(
            flowkit_transform_classname="AsinhTransform",
            feature_indices=None,
            dataset_attribute="pos",
            param_t=1,
            param_m=2,
            param_a=2,
        )
        repr(a)

    def test_AsinhTransform_repr_featureindices(self):
        a = FKtransform(
            flowkit_transform_classname="AsinhTransform",
            feature_indices=[1, 2, 3, 4],
            dataset_attribute="pos",
            param_t=1,
            param_m=2,
            param_a=2,
        )
        repr(a)

    def test_AsinhTransform(self):
        mymeta = gen_foldering_meta(
            os.path.join(TESTS_DIR, pointcloud_toy_dataset, "raw")
        )
        complete_data_notransform = InMemoryPointCloud(mymeta, clear_processed=True)
        # print(complete_data_notransform)
        # print(summary(complete_data_notransform.pos))

        complete_data = InMemoryPointCloud(
            mymeta,
            pre_transform=FKtransform(
                flowkit_transform_classname="AsinhTransform",
                feature_indices=[13],
                dataset_attribute="pos",
                param_t=1,
                param_m=2,
                param_a=2,
            ),
            clear_processed=True,
        )
        # print(summary(complete_data.pos))
        assert (
            complete_data_notransform.pos[:, 13].max() > complete_data.pos[:, 13].max()
        )

        complete_data_allfeatures = InMemoryPointCloud(
            mymeta,
            pre_transform=FKtransform(
                flowkit_transform_classname="AsinhTransform",
                param_t=1,
                param_m=2,
                param_a=2,
            ),
            clear_processed=True,
        )
        # print(summary(complete_data_allfeatures.pos))
        allfeatures_max = complete_data_allfeatures.pos.max(0).values
        notrans_max = complete_data_notransform.pos.max(0).values
        assert all(
            [allfeatures_max[i] < notrans_max[i] for i in range(len(allfeatures_max))]
        )

    def test_RatioTransform(self):
        mymeta = gen_foldering_meta(
            os.path.join(TESTS_DIR, pointcloud_toy_dataset, "raw")
        )
        complete_data_notransform = InMemoryPointCloud(mymeta, clear_processed=True)
        print(complete_data_notransform)
        # print(summary(complete_data_notransform.pos))

        with self.assertRaises(ValueError):
            complete_data = InMemoryPointCloud(
                mymeta,
                pre_transform=FKtransform(
                    flowkit_transform_classname="RatioTransform",
                    dataset_attribute="pos",
                    param_a=1,
                    param_b=2,
                    param_c=2,
                    dim_labels=None,
                ),
                clear_processed=True,
            )
            print(complete_data)



import warnings
from typing import List, Union

import torch
from torch_geometric.data.dataset import Dataset
from torch_geometric.transforms.fixed_points import FixedPoints

from pointcyto.transforms.transforms import (
    NormalizePointCloud,
    functional_transform_adapter,
    global_mean,
    global_sd,
    normalize,
)
from pointcyto.transforms.transforms_functional import (
    arcsinh,
    scale_featurewise,
    scale_low_high_01,
)
from pointcyto.transforms.utils import standard_repr


# from https://stackoverflow.com/questions/12472338/flattening-a-list-recursively
def flatten(S: list):
    if S == []:
        return S
    if isinstance(S[0], list):
        return flatten(S[0]) + flatten(S[1:])
    return S[:1] + flatten(S[1:])


class NormalizePointCloudParam(object):
    """
    Normalization in this context means
    """

    def __init__(self, inplace: bool = False, dataset_attribute: str = "pos"):
        self.inplace = inplace
        self.attr = dataset_attribute
        self.name = "NormalizePointCloudParam"
        self.related_transform = NormalizePointCloud

        # default params
        self.mean = 0
        self.std = 1

    def __call__(self, data_list: Dataset):
        """
        Args:
            tensor (Tensor): Tensor image of size (C, H, W) to be normalized.

        Returns:
            Tensor: Normalized Tensor image.
        """

        self.mean = global_mean(data_list)
        self.std = global_sd(data_list, global_means=self.mean)

        for sample in data_list:
            setattr(
                sample,
                self.attr,
                functional_transform_adapter(
                    sample, normalize, self.attr, self.mean, self.std, self.inplace
                ),
            )

        return (
            data_list,
            {
                "name": self.name,
                "related_transform": self.related_transform,
                "param": {"mean": self.mean, "std": self.std},
            },
        )

    def __repr__(self):
        return standard_repr(self)


class FixedPointsOnList(object):
    def __init__(self, num: int, replace: bool = False, allow_duplicates: bool = False):
        self.fixedpoints = FixedPoints(
            num=num, replace=replace, allow_duplicates=allow_duplicates
        )
        self.inplace = None
        self.attr = None
        self.name = "FixedPointsOnList"
        self.related_transform = None

    def __call__(self, data_list: Dataset):
        for sample_i, sample in enumerate(data_list):
            data_list[sample_i] = self.fixedpoints(sample)

        return (
            data_list,
            {
                "name": self.name,
                "related_transform": self.related_transform,
                "param": {},
            },
        )

    def __repr__(self):
        return standard_repr(self)


class ScaleLowHigh01OnList:
    def __init__(
        self,
        samplewise_feature_shifts: List[List[Union[List, float]]] = [[[1, 100]]],
        inplace: bool = False,
        dataset_attribute: str = "pos",
    ):
        """_summary_

        Args:
            samplewise_feature_shifts (List[List[Union[List, float]]], optional):
              - A list of shifts for EACH SAMPLE. Therefore the order of samples _has_ to correspond
              to the here given samplewise shifts.
              - Each element of this list corresponds to samplewise shifts _per feature_.
                - Per feature,
                    - each element can either be
                        1)  A list with a single value, then this value will be subtracted from all
                        values for that feature
                        2)  A list of two values, then the low value will be shifted to 0 and the high value to 1.
                            This is the same as min-max transformation, just with "min" and "max" replaced
                            by the given "low" and "high" values.

                Defaults to [[1, 100]] -->
                    For all samples,
                        and all features,
                            1   -> 0
                            100 -> 1
                            by
                            (x - 1)/(100-1)

            inplace (bool, optional): _description_. Defaults to False.
            dataset_attribute (str, optional): _description_. Defaults to "pos".
        """
        self.samplewise_feature_shifts = samplewise_feature_shifts
        self.name = "ScaleLowHigh01OnList"
        self.related_transform = None
        self.inplace = inplace
        self.attr = dataset_attribute
        self.related_transform = None

    def __call__(self, data_list: Dataset):
        """
        Args:
            tensor (Tensor): Tensor image of size (C, H, W) to be normalized.

        Returns:
            Tensor: Normalized Tensor image.
        """
        samplewise_feature_shifts = self.samplewise_feature_shifts
        if len(samplewise_feature_shifts) != len(data_list):
            if len(samplewise_feature_shifts) == 1:
                samplewise_feature_shifts = samplewise_feature_shifts * len(data_list)
            else:
                ValueError(
                    "Either give the same number of shifts as samples, or only a single one."
                )

        for sample_i, sample in enumerate(data_list):
            setattr(
                sample,
                self.attr,
                functional_transform_adapter(
                    sample,
                    scale_low_high_01,
                    self.attr,
                    samplewise_feature_shifts[sample_i],
                    self.inplace,
                ),
            )

        return (
            data_list,
            {
                "name": self.name,
                "related_transform": self.related_transform,
                "param": {"samplewise_feature_shifts": self.samplewise_feature_shifts},
            },
        )

    def __repr__(self):
        return standard_repr(self)


class ReTransformShiftAsinhOnList(ScaleLowHigh01OnList):
    def __init__(
        self,
        samplewise_feature_shifts_asinhspace: List[List[Union[List, float]]] = [
            [[1, 100]]
        ],
        do_shift_in_rawspace: bool = False,
        original_asinh_cofactors: List[float] = [150],
        original_asinh_cofactors_applied: bool = False,
        new_asinh_cofactors: List[float] = [10],
        inplace: bool = False,
        dataset_attribute: str = "pos",
    ):
        """_summary_

        Args:
            samplewise_feature_shifts_asinhspace (List[List[Union[List, float]]], optional):
              - A list of shifts for EACH SAMPLE. Therefore the order of samples _has_ to correspond
              to the here given samplewise shifts.
              - Each element of this list corresponds to samplewise shifts _per feature_.
                - Per feature,
                    - each element can either be
                        1)  A list with a single value, then this value will be subtracted from all
                        values for that feature
                        2)  A list of two values, then the low value will be shifted to 0 and the high value to 1.
                            This is the same as min-max transformation, just with "min" and "max" replaced
                            by the given "low" and "high" values.

                Defaults to [[1, 100]] -->
                    For all samples,
                        and all features,
                            1   -> 0
                            100 -> 1
                            by
                            (x - 1)/(100-1)
            do_shift_in_rawspace:
                Should the inversed shifts be applied in the raw space (True)?
                If False, the inversed shifts are (properly) applied in the new_asinh_cofactor-space.
            original_asinh_cofactors:
                Each element of this list corresponds to the cofactor for the asinh:
                    asinh(x_i/cofactor_i)
                per feature i which the data have been transformed with to
                identify samplewise_feature_shifts_asinhspace.
            original_asinh_cofactors_applied:
                Are the original_asinh_cofactors still applied on the data (True) or are we loading
                raw data(False)?
            new_asinh_cofactors:
                Each element of this list corresponds to the cofactor for the asinh:
                    asinh(x_i/cofactor_i)
                per feature i which the data SHOULD be transformed with.
            inplace (bool, optional): _description_. Defaults to False.
            dataset_attribute (str, optional): _description_. Defaults to "pos".
        """

        self.samplewise_feature_shifts_asinhspace = samplewise_feature_shifts_asinhspace
        self.original_asinh_cofactors = original_asinh_cofactors
        self.original_asinh_cofactors_applied = original_asinh_cofactors_applied
        self.do_shift_in_rawspace = do_shift_in_rawspace
        self.new_asinh_cofactors = new_asinh_cofactors
        self.name = "ReTransformShiftAsinhOnList"
        self.related_transform = None
        self.inplace = inplace
        self.attr = dataset_attribute
        self.related_transform = None

        if not all(
            [
                len(self.original_asinh_cofactors) == len(x)
                for x in self.samplewise_feature_shifts_asinhspace
            ]
        ):
            raise ValueError(
                "len(samplewise_feature_shifts_asinhspace) must be identical to len(original_asinh_cofactors)"
            )

        if len(self.new_asinh_cofactors) != len(self.original_asinh_cofactors):
            raise ValueError("New and old asinh cofactors must be the same length.")

        if original_asinh_cofactors_applied:
            warnings.warn(
                "Do not use this transformation on transformed AND shifted values, only on "
                + "transformed and NOT shifted."
            )
            # In theory, I have to inverse with sinh() and the original cofactors, THEN
            # apply the new transformation as is right now.
            # raise NotImplementedError

        def recursive_apply(item, fun=torch.sinh):
            if isinstance(item, list):
                return [recursive_apply(x, fun) for x in item]
            else:
                return fun(item)

        # To identify the samplewise feature shifts (in asinh-space), we did the following:
        #   1. Apply the cofactor per feature (=divide through cofactor)
        #   2. Asinh(-) transformation of all values
        #   3. Manual identification of the shifts
        # Therefore, to inverse:
        #   1. "Un-apply" the asinh (=sinh) to the manually identified shifts
        #   2. "Un-apply" the cofactor by multiplying it per feature

        # 1. "Un-apply" (inverse) the asinh
        pre_cofactor_shifts = recursive_apply(
            samplewise_feature_shifts_asinhspace,
            lambda x: torch.sinh(torch.tensor(x) * 1.0),
        )
        if any([torch.isinf(x) for x in flatten(pre_cofactor_shifts)]):
            raise ValueError(
                "torch.sinh() resulted in atleast one inf value, are your asinhspace feature shifts reasonable?"
            )

        # 2. "Un-apply" the cofactors:
        #   Go through all samples, multiply all manually idenfied shifts with the corresponding cofactors
        rawspace_shifts = []
        newspace_shifts = []
        for sample in pre_cofactor_shifts:
            rawspace_shifts.append([])
            newspace_shifts.append([])
            for feature_i, feature_shift in enumerate(sample):
                rawspace_shifts[-1].append(
                    recursive_apply(
                        feature_shift, lambda x: x * original_asinh_cofactors[feature_i]
                    )
                )
                newspace_shifts[-1].append(
                    recursive_apply(
                        rawspace_shifts[-1][-1],
                        lambda x: arcsinh(x / new_asinh_cofactors[feature_i]),
                    )
                )
        self.rawspace_shifts = rawspace_shifts
        self.newspace_shifts = newspace_shifts

    def __call__(self, data_list: Dataset):
        # samplewise_feature_shifts = self.samplewise_feature_shifts
        # if len(samplewise_feature_shifts) != len(data_list):
        #     if len(samplewise_feature_shifts) == 1:
        #         samplewise_feature_shifts = samplewise_feature_shifts * len(data_list)
        #     else:
        #         ValueError(
        #             "Either give the same number of shifts as samples, or only a single one."
        #         )
        if self.do_shift_in_rawspace:
            samplewise_feature_shifts = self.rawspace_shifts
        else:
            samplewise_feature_shifts = self.newspace_shifts
        if len(samplewise_feature_shifts) != len(data_list):
            if len(samplewise_feature_shifts) == 1:
                samplewise_feature_shifts = samplewise_feature_shifts * len(data_list)
            else:
                ValueError(
                    "Either give the same number of shifts as samples, or only a single one."
                )
        for sample_i, sample in enumerate(data_list):
            if self.original_asinh_cofactors_applied:
                # 1. Undo the asinh transformation
                setattr(
                    sample,
                    self.attr,
                    functional_transform_adapter(sample, torch.sinh, self.attr),
                )
                # 2. Undo the original feature scalings
                setattr(
                    sample,
                    self.attr,
                    functional_transform_adapter(
                        sample,
                        scale_featurewise,
                        self.attr,
                        [1 / x for x in self.original_asinh_cofactors],
                        self.inplace,
                    ),
                )
            if self.do_shift_in_rawspace:
                setattr(
                    sample,
                    self.attr,
                    functional_transform_adapter(
                        sample,
                        scale_low_high_01,
                        self.attr,
                        samplewise_feature_shifts[sample_i],
                        self.inplace,
                    ),
                )
            # 1. Apply new cofactors
            setattr(
                sample,
                self.attr,
                functional_transform_adapter(
                    sample,
                    scale_featurewise,
                    self.attr,
                    self.new_asinh_cofactors,
                    self.inplace,
                ),
            )
            # 2. Apply asinh
            setattr(
                sample,
                self.attr,
                functional_transform_adapter(sample, arcsinh, self.attr),
            )
            if not self.do_shift_in_rawspace:
                setattr(
                    sample,
                    self.attr,
                    functional_transform_adapter(
                        sample,
                        scale_low_high_01,
                        self.attr,
                        samplewise_feature_shifts[sample_i],
                        self.inplace,
                    ),
                )
        return (
            data_list,
            {
                "name": self.name,
                "related_transform": self.related_transform,
                "param": {
                    "samplewise_feature_shifts_asinhspace": self.samplewise_feature_shifts_asinhspace,
                    "do_shift_in_rawspace": self.do_shift_in_rawspace,
                    "original_asinh_cofactors": self.original_asinh_cofactors,
                    "original_asinh_cofactors_applied": self.original_asinh_cofactors_applied,
                    "new_asinh_cofactors": self.new_asinh_cofactors,
                    # "inplace": self.inplace,
                    # "dataset_attribute": self.dataset_attribute,
                },
            },
        )

    def __repr__(self):
        return standard_repr(self)



import numpy as np
from matplotlib import colormaps


# from https://gist.github.com/jakevdp/91077b0cae40f8f8244a
def discrete_cmap(n, base_cmap=None, alpha: float = 0.2, return_color_list=False):
    """
    Create an N-bin discrete colormap from the specified input map

    Args:
        n:
            How many colors do you want
        base_cmap:
            The base color map
        alpha:
            alpha from RGBA model https://en.wikipedia.org/wiki/RGBA_color_model
        return_color_list:
            Should a color list or a colormap be returned? Default: colormap

    Returns:
        Depending on return_color_list:
            - A list of colors
            - A colormap
    """

    # Note that if base_cmap is a string or None, you can simply do
    #    return plt.cm.get_cmap(base_cmap, N)
    # The following works for string, None, or a colormap instance:

    base = colormaps[base_cmap]
    color_list = base(np.linspace(0, 1, n))
    color_list[:, 3] = alpha
    if return_color_list:
        retval = color_list
    else:
        cmap_name = base.name + str(n)
        retval = base.from_list(cmap_name, color_list, n, gamma=0.3)
    return retval



import os


def stitch_path(path_or_list, relative_path: str = None) -> str:
    """
    Stitch a list of strings to a path.

    Args:
        path_or_list:
            Either a direct path or a list where all elements are stitched together to a path
        relative_path:
            If given, this path gets in the beginning of the stitched path.

    Returns:
        A string with the combined (relative_path/path_or_list_stitched)
    """
    stitched_path = (
        os.path.join(*list(path_or_list))
        if not isinstance(path_or_list, str)
        else path_or_list
    )

    if not (relative_path is None):
        if not isinstance(relative_path, str):
            relative_path = os.path.join(*relative_path)

        stitched_path = os.path.join(relative_path, stitched_path)
    return stitched_path
